/*
 *  sys_amiga.cpp - System dependent routines, Amiga implementation
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include <exec/types.h>
#include <exec/memory.h>
#include <devices/trackdisk.h>
#include <devices/scsidisk.h>
#include <resources/disk.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/disk.h>

#include "sysdeps.h"
#include "main.h"
#include "prefs.h"
#include "user_strings.h"
#include "sys.h"

#define DEBUG 0
#include "debug.h"


// File handles are pointers to these structures
struct file_handle {
	bool is_file;			// Flag: plain file or /dev/something?
	bool read_only;			// Copy of Sys_open() flag
	ULONG start_byte;		// Size of file header (if any)
	ULONG size;				// Size of file (minus header)

	BPTR f;					// AmigaDOS file handle (if is_file == true)

	struct IOStdReq *io;	// Pointer to IORequest (if is_file == false)
	ULONG block_size;		// Block size of device (must be a power of two)
};


// FileInfoBlock (must be global because it has to be on a longword boundary)
static struct FileInfoBlock FIB;

// Message port for device communication
static struct MsgPort *the_port = NULL;

// Temporary buffer in chip memory
const int TMP_BUF_SIZE = 0x10000;
static UBYTE *tmp_buf = NULL;


/*
 *  Initialization
 */

void SysInit(void)
{
	// Create port and temporary buffer
	the_port = CreateMsgPort();
	tmp_buf = (UBYTE *)AllocMem(TMP_BUF_SIZE, MEMF_CHIP | MEMF_PUBLIC);
	if (the_port == NULL || tmp_buf == NULL) {
		ErrorAlert(GetString(STR_NO_MEM_ERR));
		QuitEmulator();
	}
}


/*
 *  Deinitialization
 */

void SysExit(void)
{
	// Delete port and temporary buffer
	if (the_port)
		DeleteMsgPort(the_port);
	if (tmp_buf)
		FreeMem(tmp_buf, TMP_BUF_SIZE);
}


/*
 *  This gets called when no "floppy" prefs items are found
 *  It scans for available floppy drives and adds appropriate prefs items
 */

void SysAddFloppyPrefs(void)
{
#if 0
	for (int i=0; i<4; i++) {
		ULONG id = GetUnitID(i);
		if (id == DRT_150RPM) {	// We need an HD drive
			char str[256];
			sprintf(str, "/dev/mfm.device/%d/0/0/1474560/512", i);
			PrefsAddString("floppy", str);
		}
	}
#endif
}


/*
 *  This gets called when no "disk" prefs items are found
 *  It scans for available HFS volumes and adds appropriate prefs items
 */

void SysAddDiskPrefs(void)
{
	// AmigaOS doesn't support MacOS partitioning, so this probably doesn't make much sense...
}


/*
 *  This gets called when no "cdrom" prefs items are found
 *  It scans for available CD-ROM drives and adds appropriate prefs items
 */

void SysAddCDROMPrefs(void)
{
	// Don't scan for drives if nocdrom option given
	if (PrefsFindBool("nocdrom"))
		return;

	//!!
}


/*
 *  Add default serial prefs (must be added, even if no ports present)
 */

void SysAddSerialPrefs(void)
{
	PrefsAddString("seriala", "serial.device/0");
	PrefsAddString("serialb", "*parallel.device/0");
}


/*
 *  Open file/device, create new file handle (returns NULL on error)
 *
 *  Format for device names: /dev/<name>/<unit>/<open flags>/<start byte>/<size>/<block size>
 */

void *Sys_open(const char *name, bool read_only)
{
	bool is_file = (strstr(name, "/dev/") != name);

	D(bug("Sys_open(%s, %s)\n", name, read_only ? "read-only" : "read/write"));

	// File or device?
	if (is_file) {

		// File, open it and get stats
		BPTR f = Open((char *)name, MODE_OLDFILE);
		if (!f)
			return NULL;
		if (!ExamineFH(f, &FIB)) {
			Close(f);
			return NULL;
		}

		// Check if file is write protected
		if (FIB.fib_Protection & FIBF_WRITE)
			read_only = true;

		// Create file_handle
		file_handle *fh = new file_handle;
		fh->f = f;
		fh->is_file = true;
		fh->read_only = read_only;

		// Detect size and header length: if the file size is not a multiple of 512,
		// the remainder at the beginning is considered to be the header and skipped
		fh->size = FIB.fib_Size;
		if (FIB.fib_Size == 838484)	// 800K DiskCopy image
			fh->start_byte = 84;
		else
			fh->start_byte = FIB.fib_Size % 512;
		fh->size -= fh->start_byte;
		return fh;

	} else {

		// Device, parse string
		char dev_name[256];
		ULONG dev_unit, dev_flags, dev_start, dev_size, dev_bsize;
		if (sscanf(name, "/dev/%[^/]/%ld/%ld/%ld/%ld/%ld", dev_name, &dev_unit, &dev_flags, &dev_start, &dev_size, &dev_bsize) != 6)
			return NULL;

		// Create IORequest
		struct IOStdReq *io = (struct IOStdReq *)CreateIORequest(the_port, sizeof(struct IOExtTD));
		if (io == NULL)
			return NULL;

		// Open device
		if (OpenDevice((UBYTE *)dev_name, dev_unit, (struct IORequest *)io, dev_flags)) {
			D(bug(" couldn't open device\n"));
			DeleteIORequest(io);
			return NULL;
		}

		// Create file_handle
		file_handle *fh = new file_handle;
		fh->io = io;
		fh->is_file = false;
		fh->read_only = read_only;
		fh->start_byte = dev_start;
		fh->size = dev_size;
		fh->block_size = dev_bsize;

		return fh;
	}
}


/*
 *  Close file/device, delete file handle
 */

void Sys_close(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	D(bug("Sys_close(%08lx)\n", arg));

	// File or device?
	if (fh->is_file) {

		// File, simply close it
		Close(fh->f);

	} else {

		// Device, close it and delete IORequest
		fh->io->io_Command = CMD_UPDATE;
		DoIO((struct IORequest *)fh->io);

		fh->io->io_Command = TD_MOTOR;
		fh->io->io_Length = 0;
		DoIO((struct IORequest *)fh->io);

		CloseDevice((struct IORequest *)fh->io);
		DeleteIORequest(fh->io);
	}
	delete fh;
}


/*
 *  Read "length" bytes from file/device, starting at "offset", to "buffer",
 *  returns number of bytes read (or 0)
 */

size_t Sys_read(void *arg, void *buffer, off_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return 0;

	// File or device?
	if (fh->is_file) {

		// File, seek to position
		if (Seek(fh->f, offset + fh->start_byte, OFFSET_BEGINNING) == -1)
			return 0;

		// Read data
		LONG actual = Read(fh->f, buffer, length);
		if (actual == -1)
			return 0;
		else
			return actual;

	} else {

		// Device, pre-read (partial read of first block) neccessary?
		off_t pos = offset + fh->start_byte;
		size_t actual = 0;
		uint32 pre_offset = pos % fh->block_size;
		if (pre_offset) {

			// Yes, read one block
			fh->io->io_Command = CMD_READ;
			fh->io->io_Length = fh->block_size;
			fh->io->io_Offset = pos - pre_offset;
			fh->io->io_Data = tmp_buf;
			if (DoIO((struct IORequest *)fh->io) || fh->io->io_Actual != fh->block_size)
				return 0;

			// Copy data to destination buffer
			size_t pre_length = fh->block_size - pre_offset;
			if (pre_length > length)
				pre_length = length;
			memcpy(buffer, tmp_buf + pre_offset, pre_length);

			// Adjust data pointers
			buffer = (uint8 *)buffer + pre_length;
			pos += pre_length;
			length -= pre_length;
			actual += pre_length;
		}

		// Main read (complete reads of middle blocks) possible?
		if (length >= fh->block_size) {

			// Yes, read blocks
			size_t main_length = length & ~(fh->block_size - 1);
			fh->io->io_Command = CMD_READ;
			fh->io->io_Length = main_length;
			fh->io->io_Offset = pos;
			fh->io->io_Data = buffer;
			if (DoIO((struct IORequest *)fh->io) || fh->io->io_Actual != main_length)
				return 0;

			// Adjust data pointers
			buffer = (uint8 *)buffer + main_length;
			pos += main_length;
			length -= main_length;
			actual += main_length;
		}

		// Post-read (partial read of last block) neccessary?
		if (length) {

			// Yes, read one block
			fh->io->io_Command = CMD_READ;
			fh->io->io_Length = fh->block_size;
			fh->io->io_Offset = pos;
			fh->io->io_Data = tmp_buf;
			if (DoIO((struct IORequest *)fh->io) || fh->io->io_Actual != fh->block_size)
				return 0;

			// Copy data to destination buffer
			memcpy(buffer, tmp_buf, length);
			actual += length;
		}

		return actual;
	}
}


/*
 *  Write "length" bytes from "buffer" to file/device, starting at "offset",
 *  returns number of bytes written (or 0)
 */

size_t Sys_write(void *arg, void *buffer, off_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return 0;

	// File or device?
	if (fh->is_file) {

		// File, seek to position if neccessary
		if (Seek(fh->f, offset + fh->start_byte, OFFSET_BEGINNING) == -1)
			return 0;

		// Write data
		LONG actual = Write(fh->f, buffer, length);
		if (actual == -1)
			return 0;
		else
			return actual;

	} else {

		// Device, write data
		fh->io->io_Command = CMD_WRITE;
		fh->io->io_Length = length;
		fh->io->io_Offset = offset + fh->start_byte;
		fh->io->io_Data = buffer;
		if (DoIO((struct IORequest *)fh->io))
			return 0;
		else
			return fh->io->io_Actual;
	}
}


/*
 *  Return size of file/device (minus header)
 */

off_t SysGetFileSize(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	return fh->size;
}


/*
 *  Eject volume (if applicable)
 */

void SysEject(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	if (!fh->is_file) {

		// Flush buffer, turn off the drive motor and eject volume
		fh->io->io_Command = CMD_UPDATE;
		DoIO((struct IORequest *)fh->io);

		fh->io->io_Command = TD_MOTOR;
		fh->io->io_Length = 0;
		DoIO((struct IORequest *)fh->io);

		fh->io->io_Command = TD_EJECT;
		fh->io->io_Length = 1;
		DoIO((struct IORequest *)fh->io);
	}
}


/*
 *  Format volume (if applicable)
 */

bool SysFormat(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return true;
}


/*
 *  Check if file/device is read-only (this includes the read-only flag on Sys_open())
 */

bool SysIsReadOnly(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	if (fh->is_file) {

		// File, return flag given to Sys_open
		return fh->read_only;

	} else {

		// Device, check write protection
		fh->io->io_Flags = IOF_QUICK;
		fh->io->io_Command = TD_PROTSTATUS;
		BeginIO((struct IORequest *)fh->io);
		if (fh->io->io_Actual)
			return true;
		else
			return fh->read_only;
	}
}


/*
 *  Check if the given file handle refers to a fixed or a removable disk
 */

bool SysIsFixedDisk(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	return true;
}


/*
 *  Check if a disk is inserted in the drive (always true for files)
 */

bool SysIsDiskInserted(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	if (fh->is_file)
		return true;
	else {

		// Check medium status
		fh->io->io_Flags = IOF_QUICK;
		fh->io->io_Command = TD_CHANGESTATE;
		BeginIO((struct IORequest *)fh->io);
		if (fh->io->io_Actual)
			return false;
		else
			return true;
	}
}


/*
 *  Prevent medium removal (if applicable)
 */

void SysPreventRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	if (!fh->is_file) {

		// Send PREVENT ALLOW MEDIUM REMOVAL SCSI command
		struct SCSICmd scsi;
		static UBYTE the_cmd[6] = {0x1e, 0, 0, 0, 1, 0};
		scsi.scsi_Length = 0;
		scsi.scsi_Command = the_cmd;
		scsi.scsi_CmdLength = 6;
		scsi.scsi_Flags = SCSIF_READ;
		scsi.scsi_Status = 0;
		fh->io->io_Data = &scsi;
		fh->io->io_Length = sizeof(scsi);
		fh->io->io_Command = HD_SCSICMD;
		DoIO((struct IORequest *)fh->io);
	}
}


/*
 *  Allow medium removal (if applicable)
 */

void SysAllowRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	if (!fh->is_file) {

		// Send PREVENT ALLOW MEDIUM REMOVAL SCSI command
		struct SCSICmd scsi;
		static UBYTE the_cmd[6] = {0x1e, 0, 0, 0, 0, 0};
		scsi.scsi_Length = 0;
		scsi.scsi_Command = the_cmd;
		scsi.scsi_CmdLength = 6;
		scsi.scsi_Flags = SCSIF_READ;
		scsi.scsi_Status = 0;
		fh->io->io_Data = &scsi;
		fh->io->io_Length = sizeof(scsi);
		fh->io->io_Command = HD_SCSICMD;
		DoIO((struct IORequest *)fh->io);
	}
}


/*
 *  Read CD-ROM TOC (binary MSF format, 804 bytes max.)
 */

bool SysCDReadTOC(void *arg, uint8 *toc)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	if (fh->is_file)
		return false;
	else {

		// Send READ TOC MSF SCSI command
		struct SCSICmd scsi;
		static UBYTE read_toc_cmd[10] = {0x43, 0x02, 0, 0, 0, 0, 0, 0x03, 0x24, 0};
		scsi.scsi_Data = (UWORD *)tmp_buf;
		scsi.scsi_Length = 804;
		scsi.scsi_Command = read_toc_cmd;
		scsi.scsi_CmdLength = 10;
		scsi.scsi_Flags = SCSIF_READ;
		scsi.scsi_Status = 0;
		fh->io->io_Data = &scsi;
		fh->io->io_Length = sizeof(scsi);
		fh->io->io_Command = HD_SCSICMD;
		if (DoIO((struct IORequest *)fh->io) || scsi.scsi_Status)
			return false;
		memcpy(toc, tmp_buf, 804);
		return true;
	}
}


/*
 *  Read CD-ROM position data (Sub-Q Channel, 16 bytes, see SCSI standard)
 */

bool SysCDGetPosition(void *arg, uint8 *pos)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Play CD audio
 */

bool SysCDPlay(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, uint8 end_m, uint8 end_s, uint8 end_f)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Pause CD audio
 */

bool SysCDPause(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Resume paused CD audio
 */

bool SysCDResume(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Stop CD audio
 */

bool SysCDStop(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Perform CD audio fast-forward/fast-reverse operation starting from specified address
 */

bool SysCDScan(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, bool reverse)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	//!!
	return false;
}


/*
 *  Set CD audio volume (0..255 each channel)
 */

void SysCDSetVolume(void *arg, uint8 left, uint8 right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	//!!
}


/*
 *  Get CD audio volume (0..255 each channel)
 */

void SysCDGetVolume(void *arg, uint8 &left, uint8 &right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	//!!
}
