/*
 *  main_unix.cpp - Startup code for Unix
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "sysdeps.h"
#include "cpu_emulation.h"
#include "sys.h"
#include "xpram.h"
#include "timer.h"
#include "sony.h"
#include "disk.h"
#include "cdrom.h"
#include "scsi.h"
#include "video.h"
#include "serial.h"
#include "ether.h"
#include "clip.h"
#include "rom_patches.h"
#include "prefs.h"
#include "user_strings.h"
#include "version.h"
#include "main.h"

#define DEBUG 1
#include "debug.h"


// Constants
const char ROM_FILE_NAME[] = "ROM";


// CPU and FPU type
int CPUType;
bool CPUIs68060;
int FPUType;


// Global variables
static bool xpram_thread_active = false;	// Flag: XPRAM watchdog installed
static pthread_t xpram_thread;				// XPRAM watchdog
static bool tick_thread_active = false;		// Flag: 60Hz thread installed
static pthread_t tick_thread;				// 60Hz thread
static pthread_mutex_t intflag_mutex;		// Mutex for InterruptFlags


// Prototypes
static void *xpram_func(void *arg);
static void *tick_func(void *arg);


/*
 *  Main program
 */

int main(void)
{
	// Initialize variables
	RAMBaseHost = NULL;
	ROMBaseHost = NULL;
	srand(time(NULL));
	tzset();

	// Print some info
	printf(GetString(STR_ABOUT_TEXT1), VERSION_MAJOR, VERSION_MINOR);
	printf(" %s\n", GetString(STR_ABOUT_TEXT2));

	// Read preferences
	PrefsInit();

	// Init system routines
	SysInit();

	// Set CPU and FPU type (UAE emulation)
	CPUType = 2;
	CPUIs68060 = false;
	FPUType = 0;

	// Create area for Mac RAM
	RAMSize = PrefsFindInt32("ramsize");
	if (RAMSize < 1024*1024) {
		WarningAlert(GetString(STR_SMALL_RAM_WARN));
		RAMSize = 1024*1024;
	}
	RAMBaseHost = new uint8[RAMSize];
#if REAL_ADDRESSING
	RAMBaseMac = (uint32)RAMBaseHost;
#else
	RAMBaseMac = 0;
#endif

	// Create area for Mac ROM
	ROMBaseHost = new uint8[0x100000];
#if REAL_ADDRESSING
	ROMBaseMac = (uint32)ROMBaseHost;
#else
	ROMBaseMac = 0x40800000;
#endif

	// Get rom file path from preferences
	const char *rom_path = PrefsFindString("rom");

	// Load Mac ROM
	int rom_fd = open(rom_path ? rom_path : ROM_FILE_NAME, O_RDONLY);
	if (rom_fd < 0) {
		ErrorAlert(GetString(STR_NO_ROM_FILE_ERR));
		QuitEmulator();
	}
	printf(GetString(STR_READING_ROM_FILE));
	ROMSize = lseek(rom_fd, 0, SEEK_END);
	if (ROMSize != 1024*1024 && ROMSize != 512*1024) {
		ErrorAlert(GetString(STR_ROM_SIZE_ERR));
		close(rom_fd);
		QuitEmulator();
	}
	lseek(rom_fd, 0, SEEK_SET);
	if (read(rom_fd, ROMBaseHost, ROMSize) != (ssize_t)ROMSize) {
		ErrorAlert(GetString(STR_ROM_FILE_READ_ERR));
		close(rom_fd);
		QuitEmulator();
	}

	// Load XPRAM
	XPRAMInit();

	// Set boot volume
	int16 i16 = PrefsFindInt16("bootdrive");
	XPRAM[0x78] = i16 >> 8;
	XPRAM[0x79] = i16 & 0xff;
	i16 = PrefsFindInt16("bootdriver");
	XPRAM[0x7a] = i16 >> 8;
	XPRAM[0x7b] = i16 & 0xff;

	// Init InterruptFlags mutex
	pthread_mutex_init(&intflag_mutex, NULL);

	// Start XPRAM watchdog thread
	xpram_thread_active = (pthread_create(&xpram_thread, NULL, xpram_func, NULL) == 0);

	// Init drivers
	SonyInit();
	DiskInit();
	CDROMInit();
	SCSIInit();

	// Init serial ports
	SerialInit();

	// Init network
	EtherInit();

	// Init Time Manager
	TimerInit();

	// Init clipboard
	ClipInit();

	// Init video
	if (!VideoInit())
		QuitEmulator();

	// Init 680x0 emulation (this also activates the memory system which is needed for PatchROM())
	if (!Init680x0())
		QuitEmulator();

	// Install ROM patches
	if (!PatchROM()) {
		ErrorAlert(GetString(STR_UNSUPPORTED_ROM_TYPE_ERR));
		QuitEmulator();
	}

	// Start 60Hz thread
	tick_thread_active = (pthread_create(&tick_thread, NULL, tick_func, NULL) == 0);

	// Start 68k and jump to ROM boot routine
	Start680x0();

	QuitEmulator();
	return 0;
}


/*
 *  Quit emulator
 */

void QuitEmulator(void)
{
	// Exit 680x0 emulation
	Exit680x0();

	// Stop 60Hz thread
	if (tick_thread_active) {
		pthread_cancel(tick_thread);
		pthread_join(tick_thread, NULL);
	}

	// Stop XPRAM watchdog thread
	if (xpram_thread_active) {
		pthread_cancel(xpram_thread);
		pthread_join(xpram_thread, NULL);
	}

	// Delete InterruptFlags mutex
	pthread_mutex_destroy(&intflag_mutex);

	// Save XPRAM
	XPRAMExit();

	// Exit video
	VideoExit();

	// Exit clipboard
	ClipExit();

	// Exit Time Manager
	TimerExit();

	// Exit serial ports
	SerialExit();

	// Exit network
	EtherExit();

	// Exit drivers
	SCSIExit();
	CDROMExit();
	DiskExit();
	SonyExit();

	// Delete ROM area
	delete[] ROMBaseHost;

	// Delete RAM area
	delete[] RAMBaseHost;

	// Exit system routines
	SysExit();

	// Exit preferences
	PrefsExit();

	exit(0);
}


/*
 *  Code was patched, flush caches if neccessary (i.e. when using a real 680x0
 *  or a dynamically recompiling emulator)
 */

#if EMULATED_68K
void FlushCodeCache(void *start, uint32 size)
{
}
#endif


/*
 *  Interrupt flags (must be handled atomically!)
 */

uint32 InterruptFlags = 0;

void SetInterruptFlag(uint32 flag)
{
	pthread_mutex_lock(&intflag_mutex);
	InterruptFlags |= flag;
	pthread_mutex_unlock(&intflag_mutex);
}

void ClearInterruptFlag(uint32 flag)
{
	pthread_mutex_lock(&intflag_mutex);
	InterruptFlags &= ~flag;
	pthread_mutex_unlock(&intflag_mutex);
}


/*
 *  60Hz thread (really 60.15Hz)
 */

// Mac time starts in 1904, Unix time in 1970, this is the offset in seconds
const uint32 TIME_OFFSET = 0x7c25b080;

static void *tick_func(void *arg)
{
	int tick_counter = 0;
	timeval t;

	for (;;) {

		// Wait
		pthread_testcancel();
		usleep(16625);
		pthread_testcancel();

		// Pseudo Mac 1Hz interrupt, update local time
		if (++tick_counter > 60) {
			tick_counter = 0;
			gettimeofday(&t, NULL);
			WriteMacInt32(0x20c, t.tv_sec - timezone + (daylight ? 3600 : 0) + TIME_OFFSET);
		}

		// Trigger 60Hz interrupt
		SetInterruptFlag(INTFLAG_60HZ);
		TriggerInterrupt();
	}
	return NULL;
}


/*
 *  XPRAM watchdog thread (saves XPRAM every minute)
 */

void *xpram_func(void *arg)
{
	uint8 last_xpram[256];
	memcpy(last_xpram, XPRAM, 256);

	for (;;) {
		pthread_testcancel();
		sleep(60);
		pthread_testcancel();
		if (memcmp(last_xpram, XPRAM, 256)) {
			memcpy(last_xpram, XPRAM, 256);
			SaveXPRAM();
		}
	}
	return NULL;
}


/*
 *  Display error alert
 */

void ErrorAlert(const char *text)
{
	printf(GetString(STR_SHELL_ERROR_PREFIX), text);
}


/*
 *  Display warning alert
 */

void WarningAlert(const char *text)
{
	printf(GetString(STR_SHELL_WARNING_PREFIX), text);
}


/*
 *  Display choice alert
 */

bool ChoiceAlert(const char *text, const char *pos, const char *neg)
{
	printf(GetString(STR_SHELL_WARNING_PREFIX), text);
	return false;	//!!
}
