/*
 *  video_x.cpp - Video/graphics emulation, X11 specific stuff
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pthread.h>

#include "sysdeps.h"
#include "cpu_emulation.h"
#include "main.h"
#include "adb.h"
#include "prefs.h"
#include "user_strings.h"
#include "video.h"

#define DEBUG 0
#include "debug.h"

#if ENABLE_DGA
#include <X11/extensions/xf86dga.h>
#endif


// Display types
enum {
	DISPLAY_WINDOW,	// X11 window, using MIT SHM extensions
	DISPLAY_DGA		// DGA fullscreen display
};


// Global variables
static int32 frame_skip;
static int display_type = DISPLAY_WINDOW;	// See enum above
static uint8 *the_buffer;					// Mac frame buffer
static bool redraw_thread_active = false;	// Flag: Redraw thread installed
static pthread_t redraw_thread;				// Redraw thread

Display *x_display = NULL;
static int screen;
static Window rootwin, the_win;
static XVisualInfo visualInfo;
static Visual *vis;
static Colormap cmap[2];
static XColor black, white;
static unsigned long black_pixel, white_pixel;
static const int eventmask = KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask | EnterWindowMask;

// Variables for window mode
static GC black_gc;
static XImage *img = NULL;
static XShmSegmentInfo shminfo;
static XImage *cursor_image, *cursor_mask_image;
static Pixmap cursor_map, cursor_mask_map;
static Cursor mac_cursor;
static GC cursor_gc, cursor_mask_gc;
static uint8 the_cursor[64];				// Cursor image data

// Variables for DGA mode
static int current_dga_cmap;


// Prototypes
static void *redraw_func(void *arg);
static int event2keycode(XKeyEvent *ev);


/*
 *  Initialization
 */

// Set VideoMonitor according to video mode
void set_video_monitor(int width, int height, int depth, int bytes_per_row)
{
	switch (depth) {
		case 8:
			MacFrameLayout = FLAYOUT_DIRECT;
			VideoMonitor.mode = VMODE_8BIT;
			break;
		case 15:
			MacFrameLayout = FLAYOUT_HOST_555;
			VideoMonitor.mode = VMODE_16BIT;
			break;
		case 16:
			MacFrameLayout = FLAYOUT_HOST_565;
			VideoMonitor.mode = VMODE_16BIT;
			break;
		case 24:
		case 32:
			MacFrameLayout = FLAYOUT_HOST_888;
			VideoMonitor.mode = VMODE_32BIT;
			break;
	}
	VideoMonitor.x = width;
	VideoMonitor.y = height;
	VideoMonitor.bytes_per_row = bytes_per_row;
}

// Init window mode
static bool init_window(int width, int height, int depth)
{
	// Set absolute mouse mode
	ADBSetRelMouseMode(false);
	BlankMousePointer = true;

	// Read frame skip prefs
	frame_skip = PrefsFindInt32("frameskip");
	if (frame_skip == 0)
		frame_skip = 1;

	// Create color maps for 8 bit mode
	if (depth == 8) {
		cmap[0] = XCreateColormap(x_display, rootwin, vis, AllocAll);
		cmap[1] = XCreateColormap(x_display, rootwin, vis, AllocAll);
	}

	// Create window
	XSetWindowAttributes wattr;
	wattr.event_mask = eventmask;
	wattr.background_pixel = black_pixel;
	wattr.border_pixel = black_pixel;
	wattr.backing_store = Always;
	wattr.backing_planes = depth;
	wattr.colormap = cmap[0];

	the_win = XCreateWindow(x_display, rootwin, 0, 0, width, height, 0, depth,
		InputOutput, vis, CWEventMask | CWBackPixel | CWBorderPixel |
		CWBackingStore | CWBackingPlanes | (depth == 8 ? CWColormap : 0), &wattr);
	XMapWindow(x_display, the_win);
	XStoreName(x_display, the_win, GetString(STR_WINDOW_TITLE));

	// Make window unresizable
	XSizeHints *hints;
	if ((hints = XAllocSizeHints()) != NULL) {
		hints->min_width = width;
		hints->max_width = width;
		hints->min_height = height;
		hints->max_height = height;
		hints->flags = PMinSize | PMaxSize;
		XSetWMNormalHints(x_display, the_win, hints);
		XFree((char *)hints);
	}

	// Create and attach SHM image ("height + 2" for safety)
	img = XShmCreateImage(x_display, vis, depth, ZPixmap, 0, &shminfo, width, height);
	shminfo.shmid = shmget(IPC_PRIVATE, (height + 2) * img->bytes_per_line, IPC_CREAT | 0777);
	the_buffer = (uint8 *)shmat(shminfo.shmid, 0, 0);
	shminfo.shmaddr = img->data = (char *)the_buffer;
	shminfo.readOnly = False;
	XShmAttach(x_display, &shminfo);
	XSync(x_display, false);
	shmctl(shminfo.shmid, IPC_RMID, 0);

	// Create GC
	black_gc = XCreateGC(x_display, the_win, 0, 0);
	XSetForeground(x_display, black_gc, black_pixel);

	// Create cursor
	cursor_image = XCreateImage(x_display, vis, 1, XYPixmap, 0, (char *)the_cursor, 16, 16, 16, 2);
	cursor_image->byte_order = MSBFirst;
	cursor_image->bitmap_bit_order = MSBFirst;
	cursor_mask_image = XCreateImage(x_display, vis, 1, XYPixmap, 0, (char *)the_cursor+32, 16, 16, 16, 2);
	cursor_mask_image->byte_order = MSBFirst;
	cursor_mask_image->bitmap_bit_order = MSBFirst;
	cursor_map = XCreatePixmap(x_display, the_win, 16, 16, 1);
	cursor_mask_map = XCreatePixmap(x_display, the_win, 16, 16, 1);
	cursor_gc = XCreateGC(x_display, cursor_map, 0, 0);
	cursor_mask_gc = XCreateGC(x_display, cursor_mask_map, 0, 0);
	mac_cursor = XCreatePixmapCursor(x_display, cursor_map, cursor_mask_map, &black, &white, 0, 0);

	// Set VideoMonitor
	set_video_monitor(width, height, depth, img->bytes_per_line);
#if REAL_ADDRESSING
	VideoMonitor.mac_frame_base = (uint32)the_buffer;
	MacFrameLayout = FLAYOUT_DIRECT;
#else
	VideoMonitor.mac_frame_base = MacFrameBaseMac;
#endif
	return true;
}

// Init DGA display
static bool init_dga(int width, int height, int depth)
{
#if ENABLE_DGA
	// Set relative mouse mode
	ADBSetRelMouseMode(true);
	BlankMousePointer = false;

	// For 8 bit mode, create color maps
	if (depth == 8) {
		cmap[0] = XCreateColormap(x_display, rootwin, vis, AllocAll);
		cmap[1] = XCreateColormap(x_display, rootwin, vis, AllocAll);
	}

	// Create window
	XSetWindowAttributes wattr;
	wattr.event_mask = eventmask;
	wattr.background_pixel = black_pixel;
	wattr.border_pixel = black_pixel;
	wattr.colormap = cmap[0];

	the_win = XCreateWindow(x_display, rootwin, 0, 0, width, height, 0, depth,
		InputOutput, vis, CWEventMask | CWBackPixel | CWBorderPixel |
		(depth == 8 ? CWColormap : 0), &wattr);

	XMapWindow(x_display, the_win);
	XStoreName(x_display, the_win, GetString(STR_WINDOW_TITLE));

	// Establish direct screen connection
	XRaiseWindow(x_display, the_win);
	XMoveResizeWindow(x_display, the_win, 0, 0, width, height);
	XGrabKeyboard(x_display, rootwin, 1, GrabModeAsync, GrabModeAsync, CurrentTime);
	XGrabPointer(x_display, rootwin, 1, PointerMotionMask | ButtonPressMask | ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

	int v_width, v_bank, v_size;
	XF86DGAGetVideo(x_display, screen, (char **)&the_buffer, &v_width, &v_bank, &v_size);
	XF86DGADirectVideo(x_display, screen, XF86DGADirectGraphics | XF86DGADirectKeyb | XF86DGADirectMouse);
	XF86DGASetViewPort(x_display, screen, 0, 0);

	if (depth == 8)
		XF86DGAInstallColormap(x_display, screen, cmap[current_dga_cmap = 0]);

	// Set VideoMonitor
	int bytes_per_row = (v_width + 7) & ~7;
	switch (depth) {
		case 15:
		case 16:
			bytes_per_row *= 2;
			break;
		case 24:
		case 32:
			bytes_per_row *= 4;
			break;
	}
	set_video_monitor(width, height, depth, bytes_per_row);
#if REAL_ADDRESSING
	VideoMonitor.mac_frame_base = (uint32)the_buffer;
	MacFrameLayout = FLAYOUT_DIRECT;
#else
	VideoMonitor.mac_frame_base = MacFrameBaseMac;
#endif
	return true;
#else
	ErrorAlert("Basilisk II has been compiled with DGA support disabled.");
	return false;
#endif
}

bool VideoInit(void)
{
	// Enable thread support
	XInitThreads();

	// Open display
	x_display = XOpenDisplay(NULL);
	if (x_display == NULL) {
		char str[256];
		sprintf(str, GetString(STR_NO_XSERVER_ERR), XDisplayName(NULL));
		ErrorAlert(str);
		return false;
	}

	// Find screen and root window
	screen = XDefaultScreen(x_display);
	rootwin = XRootWindow(x_display, screen);

	// Get screen depth
	int depth = DefaultDepth(x_display, screen);

	// Find black and white colors
	XParseColor(x_display, DefaultColormap(x_display, screen), "rgb:00/00/00", &black);
	XAllocColor(x_display, DefaultColormap(x_display, screen), &black);
	XParseColor(x_display, DefaultColormap(x_display, screen), "rgb:ff/ff/ff", &white);
	XAllocColor(x_display, DefaultColormap(x_display, screen), &white);
	black_pixel = BlackPixel(x_display, screen);
	white_pixel = WhitePixel(x_display, screen);

	// Get appropriate visual
	int color_class;
	switch (depth) {
		case 8:
			color_class = PseudoColor;
			break;
		case 15:
		case 16:
		case 24:
		case 32:
			color_class = TrueColor;
			break;
		default:
			ErrorAlert(GetString(STR_UNSUPP_DEPTH_ERR));
			return false;
	}
	if (!XMatchVisualInfo(x_display, screen, depth, color_class, &visualInfo)) {
		ErrorAlert(GetString(STR_NO_XVISUAL_ERR));
		return false;
	}
	if (visualInfo.depth != depth) {
		ErrorAlert(GetString(STR_NO_XVISUAL_ERR));
		return false;
	}
	vis = visualInfo.visual;

	// Get screen mode from preferences
	const char *mode_str = PrefsFindString("screen");

	// Determine type and mode
	display_type = DISPLAY_WINDOW;
	int width = 512, height = 384;
	if (mode_str) {
		if (sscanf(mode_str, "win/%d/%d", &width, &height) == 2)
			display_type = DISPLAY_WINDOW;
		else if (strcmp(mode_str, "dga") == 0) {
			display_type = DISPLAY_DGA;
			width = DisplayWidth(x_display, screen);
			height = DisplayHeight(x_display, screen);
		}
	}

	// Initialize according to display type
	switch (display_type) {
		case DISPLAY_WINDOW:
			if (!init_window(width, height, depth))
				return false;
			break;
		case DISPLAY_DGA:
			if (!init_dga(width, height, depth))
				return false;
			break;
	}

#if !REAL_ADDRESSING
	// Set variables for UAE memory mapping
	MacFrameBaseHost = the_buffer;
	MacFrameSize = VideoMonitor.bytes_per_row * VideoMonitor.y;
#endif

	// Start redraw thread for window mode
	XSync(x_display, false);
	if (display_type == DISPLAY_WINDOW)
		redraw_thread_active = (pthread_create(&redraw_thread, NULL, redraw_func, NULL) == 0);
	return true;
}


/*
 *  Deinitialization
 */

void VideoExit(void)
{
	// Stop redraw thread
	if (redraw_thread_active) {
		pthread_cancel(redraw_thread);
		pthread_join(redraw_thread, NULL);
	}

	// Close window and server connection
	if (x_display != NULL) {
		XSync(x_display, false);

#if ENABLE_DGA
		if (display_type == DISPLAY_DGA) {
			XF86DGADirectVideo(x_display, screen, 0);
			XUngrabPointer(x_display, CurrentTime);
			XUngrabKeyboard(x_display, CurrentTime);
		}
#endif

		XCloseDisplay(x_display);
	}
}


/*
 *  Set palette
 */

void video_set_palette(uint8 *pal)
{
	// Convert colors to XColor array
	XColor c[256];
	for (int i=0; i<256; i++) {
		c[i].pixel = i;
		c[i].red = pal[i*3] * 0x0101;
		c[i].green = pal[i*3+1] * 0x0101;
		c[i].blue = pal[i*3+2] * 0x0101;
		c[i].flags = DoRed | DoGreen | DoBlue;
	}

	// Lock X server access
	XLockDisplay(x_display);

	// Set colormaps
	XStoreColors(x_display, cmap[0], c, 256);
	XStoreColors(x_display, cmap[1], c, 256);

#if ENABLE_DGA
	if (display_type == DISPLAY_DGA) {
		current_dga_cmap ^= 1;
		XF86DGAInstallColormap(x_display, screen, cmap[current_dga_cmap]);
	}
#endif

	// Unlock X server access
	XUnlockDisplay(x_display);
}


/*
 *  Thread for screen refresh
 */

static void *redraw_func(void *arg)
{
	int tick_counter = 0;

	for (;;) {

		// Wait
		pthread_testcancel();
		usleep(1000000 / 60);
		pthread_testcancel();

		// Lock X server access
		XLockDisplay(x_display);

		tick_counter++;
		if (tick_counter >= frame_skip) {
			tick_counter = 0;

			// Refresh display
			XShmPutImage(x_display, the_win, black_gc, img, 0, 0, 0, 0, VideoMonitor.x, VideoMonitor.y, 0);

			// Has the Mac started?
			if (ReadMacInt32(0xcfc) == 'WLSC') {

				// Set new cursor image if it was changed
				if (memcmp(the_cursor, Mac2HostAddr(0x844), 64)) {
					memcpy(the_cursor, Mac2HostAddr(0x844), 64);
					memcpy(cursor_image->data, the_cursor, 32);
					memcpy(cursor_mask_image->data, the_cursor+32, 32);
					XFreeCursor(x_display, mac_cursor);
					XPutImage(x_display, cursor_map, cursor_gc, cursor_image, 0, 0, 0, 0, 16, 16);
					XPutImage(x_display, cursor_mask_map, cursor_mask_gc, cursor_mask_image, 0, 0, 0, 0, 16, 16);
					mac_cursor = XCreatePixmapCursor(x_display, cursor_map, cursor_mask_map, &black, &white, ReadMacInt8(0x885), ReadMacInt8(0x887));
					XDefineCursor(x_display, the_win, mac_cursor);
				}
			}
		}

		// Unlock X server access
		XUnlockDisplay(x_display);
	}
	return NULL;
}


/*
 *  X event handling
 */

// F12 key works like a qualifier for special actions (F12-q = emergency quit)
static bool F12_down = false;

void VideoInterrupt(void)
{
	// Lock X server access
	XLockDisplay(x_display);

	// Handle events
	for (;;) {
		XEvent event;

		if (!XCheckMaskEvent(x_display, eventmask, &event))
			break;

		switch (event.type) {
			// Mouse button
			case ButtonPress: {
				unsigned int button = ((XButtonEvent *)&event)->button;
				if (button < 4)
					ADBMouseDown(button - 1);
				break;
			}
			case ButtonRelease: {
				unsigned int button = ((XButtonEvent *)&event)->button;
				if (button < 4)
					ADBMouseUp(button - 1);
				break;
			}

			// Mouse moved
			case EnterNotify:
				ADBMouseMoved(((XMotionEvent *)&event)->x, ((XMotionEvent *)&event)->y);
				break;
			case MotionNotify:
				ADBMouseMoved(((XMotionEvent *)&event)->x, ((XMotionEvent *)&event)->y);
				break;

			// Keyboard
			case KeyPress: {
				int code;
				if ((code = event2keycode((XKeyEvent *)&event)) != -1) {
					ADBKeyDown(code);
					if (code == 0x6f)
						F12_down = true;
				}
				break;
			}
			case KeyRelease: {
				int code;
				if ((code = event2keycode((XKeyEvent *)&event)) != -1) {
					ADBKeyUp(code);
					if (code == 0x6f)
						F12_down = false;
				}
				break;
			}
		}
	}

	// Unlock X server access
	XUnlockDisplay(x_display);
}


/*
 *  Translate key event to Mac keycode
 */

static int kc_decode(KeySym ks)
{
	switch (ks) {
		case XK_A: case XK_a: return 0x00;
		case XK_B: case XK_b: return 0x0b;
		case XK_C: case XK_c: return 0x08;
		case XK_D: case XK_d: return 0x02;
		case XK_E: case XK_e: return 0x0e;
		case XK_F: case XK_f: return 0x03;
		case XK_G: case XK_g: return 0x05;
		case XK_H: case XK_h: return 0x04;
		case XK_I: case XK_i: return 0x22;
		case XK_J: case XK_j: return 0x26;
		case XK_K: case XK_k: return 0x28;
		case XK_L: case XK_l: return 0x25;
		case XK_M: case XK_m: return 0x2e;
		case XK_N: case XK_n: return 0x2d;
		case XK_O: case XK_o: return 0x1f;
		case XK_P: case XK_p: return 0x23;
		case XK_Q: case XK_q: if (F12_down) QuitEmulator(); return 0x0c;
		case XK_R: case XK_r: return 0x0f;
		case XK_S: case XK_s: return 0x01;
		case XK_T: case XK_t: return 0x11;
		case XK_U: case XK_u: return 0x20;
		case XK_V: case XK_v: return 0x09;
		case XK_W: case XK_w: return 0x0d;
		case XK_X: case XK_x: return 0x07;
		case XK_Y: case XK_y: return 0x10;
		case XK_Z: case XK_z: return 0x06;

		case XK_1: case XK_exclam: return 0x12;
		case XK_2: case XK_at: return 0x13;
		case XK_3: case XK_numbersign: return 0x14;
		case XK_4: case XK_dollar: return 0x15;
		case XK_5: case XK_percent: return 0x17;
		case XK_6: return 0x16;
		case XK_7: return 0x1a;
		case XK_8: return 0x1c;
		case XK_9: return 0x19;
		case XK_0: return 0x1d;

		case XK_grave: case XK_asciitilde: return 0x0a;
		case XK_minus: case XK_underscore: return 0x1b;
		case XK_equal: case XK_plus: return 0x18;
		case XK_bracketleft: case XK_braceleft: return 0x21;
		case XK_bracketright: case XK_braceright: return 0x1e;
		case XK_backslash: case XK_bar: return 0x2a;
		case XK_semicolon: case XK_colon: return 0x29;
		case XK_apostrophe: case XK_quotedbl: return 0x27;
		case XK_comma: case XK_less: return 0x2b;
		case XK_period: case XK_greater: return 0x2f;
		case XK_slash: case XK_question: return 0x2c;

		case XK_Tab: return 0x30;
		case XK_Return: return 0x24;
		case XK_space: return 0x31;
		case XK_BackSpace: return 0x33;

		case XK_Delete: return 0x75;
		case XK_Insert: return 0x72;
		case XK_Home: case XK_Help: return 0x73;
		case XK_End: return 0x77;
#ifdef __hpux
		case XK_Prior: return 0x74;
		case XK_Next: return 0x79;
#else
		case XK_Page_Up: return 0x74;
		case XK_Page_Down: return 0x79;
#endif

		case XK_Control_L: return 0x36;
		case XK_Control_R: return 0x36;
		case XK_Shift_L: return 0x38;
		case XK_Shift_R: return 0x38;
		case XK_Alt_L: return 0x37;
		case XK_Alt_R: return 0x37;
		case XK_Meta_L: return 0x3a;
		case XK_Meta_R: return 0x3a;
		case XK_Menu: return 0x32;
		case XK_Caps_Lock: return 0x39;
		case XK_Num_Lock: return 0x47;

		case XK_Up: return 0x3e;
		case XK_Down: return 0x3d;
		case XK_Left: return 0x3b;
		case XK_Right: return 0x3c;

		case XK_Escape: return 0x35;

		case XK_F1: return 0x7a;
		case XK_F2: return 0x78;
		case XK_F3: return 0x63;
		case XK_F4: return 0x76;
		case XK_F5: return 0x60;
		case XK_F6: return 0x61;
		case XK_F7: return 0x62;
		case XK_F8: return 0x64;
		case XK_F9: return 0x65;
		case XK_F10: return 0x6d;
		case XK_F11: return 0x67;
		case XK_F12: return 0x6f;

		case XK_Print: return 0x69;
		case XK_Scroll_Lock: return 0x6b;
		case XK_Pause: return 0x71;

#if defined(XK_KP_Prior) && defined(XK_KP_Left) && defined(XK_KP_Insert) && defined (XK_KP_End)
		case XK_KP_0: case XK_KP_Insert: return 0x52;
		case XK_KP_1: case XK_KP_End: return 0x53;
		case XK_KP_2: case XK_KP_Down: return 0x54;
		case XK_KP_3: case XK_KP_Next: return 0x55;
		case XK_KP_4: case XK_KP_Left: return 0x56;
		case XK_KP_5: case XK_KP_Begin: return 0x57;
		case XK_KP_6: case XK_KP_Right: return 0x58;
		case XK_KP_7: case XK_KP_Home: return 0x59;
		case XK_KP_8: case XK_KP_Up: return 0x5b;
		case XK_KP_9: case XK_KP_Prior: return 0x5c;
		case XK_KP_Decimal: case XK_KP_Delete: return 0x41;
#else
		case XK_KP_0: return 0x52;
		case XK_KP_1: return 0x53;
		case XK_KP_2: return 0x54;
		case XK_KP_3: return 0x55;
		case XK_KP_4: return 0x56;
		case XK_KP_5: return 0x57;
		case XK_KP_6: return 0x58;
		case XK_KP_7: return 0x59;
		case XK_KP_8: return 0x5b;
		case XK_KP_9: return 0x5c;
		case XK_KP_Decimal: return 0x41;
#endif
		case XK_KP_Add: return 0x45;
		case XK_KP_Subtract: return 0x4e;
		case XK_KP_Multiply: return 0x43;
		case XK_KP_Divide: return 0x4b;
		case XK_KP_Enter: return 0x4c;
		case XK_KP_Equal: return 0x51;
	}
	return -1;
}

static int event2keycode(XKeyEvent *ev)
{
	KeySym ks;
	int as;
	int i = 0;

	do {
		ks = XLookupKeysym(ev, i++);
		as = kc_decode(ks);
		if (as != -1)
			return as;
	} while (ks != NoSymbol);

	return -1;
}
