/*
 *  main_windows.cpp - Startup code for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include <stdio.h>

#include "sysdeps.h"
#include "uae_readcpu.h"
#include "uae_memory.h"
#include "uae_newcpu.h"
#include "sys.h"
#include "xpram.h"
#include "timer.h"
#include "sony.h"
#include "disk.h"
#include "cdrom.h"
#include "scsi.h"
#include "video.h"
#include "serial.h"
#include "ether.h"
#include "clip.h"
#include "rom_patches.h"
#include "prefs.h"
#include "user_strings.h"
#include "version.h"
#include "main.h"
#include "adb.h"
#include "main_windows.h"
#include "video_windows.h"
#include "timer_windows.h"
#include "xpram_windows.h"

#define DEBUG 0
#include "debug.h"

// Constants
const char ROM_FILE_NAME[] = "ROM";

// RAM and ROM pointers
uint32 RAMBaseMac;		// RAM base (Mac address space)
uint8 *RAMBaseHost;		// RAM base (host address space)
uint32 RAMSize;				// Size of RAM
int32 RAMBaseDiff;		// 

uint32 ROMBaseMac;		// ROM base (Mac address space)
uint8 *ROMBaseHost;		// ROM base (host address space)
uint32 ROMSize;				// Size of ROM
int32 ROMBaseDiff;		// 

// CPU and FPU type
int CPUType;
bool CPUIs68060;
int FPUType;

uint32 InterruptFlags = 0;
// static HANDLE intflag_mutex = 0;

// os flag set up in initialize()
static int win_os; 

HWND hMainWnd = 0;
HINSTANCE hInst = 0;

// mouse and keyboard
static bool capturing = false;


// Keep track what Mac keys are down so we can make them up
// when we lose focus.
// This could be done easily using array in adb.cpp, but I want to keep
// changes in files outside of windows folder at minimum.
// Will ease upgrading to new versions
#define MAX_KEYS 256
static BYTE mac_keys_down[MAX_KEYS];

// Floppy detection
#define FLOPPY_KEYS_SECONDS 2
static bool is_floppy_keys_down = false;
static DWORD floppy_keys_ok_time = 0;

// Mac time starts in 1904, Unix time in 1970, this is the offset in seconds
const uint32 TIME_OFFSET = 0x7c25b080;

// For suspend/resume
static HANDLE CPUThread_handle = 0;

// local forwards
DWORD WINAPI cpu_thread(LPVOID param);
DWORD WINAPI tick_func(LPVOID param);

// #define DOWN_MASK 0x8000
#define DOWN_MASK 0x80
#define TOGGLE_MASK 1

SHORT my_get_key_state( int vKey )
{
	BYTE state[256];

	// Win95 version of GetAsyncKeyState() does not distinguish between left and right,
	// but it seems to be that even GetKeyboardState() has problems.
	// return(GetAsyncKeyState(vKey));

	GetKeyboardState( state );
	return( state[vKey] );
}

// TODO: use these as default values, read the table from a file
static int key_win2mac( WPARAM wparam, bool down )
{
	switch (wparam) {
    case 'A': return 0x00;
    case 'B': return 0x0b;
    case 'C': return 0x08;
    case 'D': return 0x02;
    case 'E': return 0x0e;
    case 'F': return 0x03;
    case 'G': return 0x05;
    case 'H': return 0x04;
    case 'I': return 0x22;
    case 'J': return 0x26;
    case 'K': return 0x28;
    case 'L': return 0x25;
    case 'M': return 0x2e;
    case 'N': return 0x2d;
    case 'O': return 0x1f;
    case 'P': return 0x23;
    case 'Q': return 0x0c;
    case 'R': return 0x0f;
    case 'S': return 0x01;
    case 'T': return 0x11;
    case 'U': return 0x20;
    case 'V': return 0x09;
    case 'W': return 0x0d;
    case 'X': return 0x07;
    case 'Y': return 0x10;
    case 'Z': return 0x06;

		case '1': return 0x12;
		case '2': return 0x13;
		case '3': return 0x14;
		case '4': return 0x15;
		case '5': return 0x17;
		case '6': return 0x16;
		case '7': return 0x1a;
		case '8': return 0x1c;
		case '9': return 0x19;
		case '0': return 0x1d;

		case VK_UP: return 0x3e;
		case VK_DOWN: return 0x3d;
		case VK_LEFT: return 0x3b;
		case VK_RIGHT: return 0x3c;

		case VK_PRIOR: return 0x74;
		case VK_NEXT: return 0x79;

		case VK_DELETE: return 0x75;
		case VK_INSERT: return 0x72;
		case VK_HOME: return 0x73;
		case VK_END: return 0x77;

		// Is it really the case that these keys don't have proper VK_* values?
		case 0xBA: return 0x1e; //  ^
		case 0xDB: return 0x18; //  `
		case 0xBB: return 0x1b; // + ?
		case 0xBC: return 0x2b; // , ;
		case 0xBD: return 0x2c; // - _
		case 0xBE: return 0x2f; // . :
		case 0xBF: return 0x2a; // ' *
		case 0xC0: return 0x29; //  
		case 0xDC: return 0xa;  //  
		case 0xDD: return 0x21; //  
		case 0xDE: return 0x27; //  
		case 0xE2: return 0x32; // < > |
		
		case VK_TAB: return 0x30;
		case VK_RETURN: return 0x24;
		case VK_SPACE: return 0x31;
		case VK_BACK: return 0x33;
		case VK_ESCAPE: return 0x35;



		// Modifier keys

		case VK_SHIFT: return 0x38;

		case VK_CONTROL: 
			// Right control maps to option
			// Don't know which one of the controls was pressed/released, find out
			bool r_is_down, l_is_down, r_was_down, l_was_down;
			l_is_down = (my_get_key_state(VK_LCONTROL) & DOWN_MASK) != 0;
			r_is_down = (my_get_key_state(VK_RCONTROL) & DOWN_MASK) != 0;
			l_was_down = mac_keys_down[0x36] != 0;
			r_was_down = mac_keys_down[0x3A] != 0;

			if(l_is_down != l_was_down) {
				return 0x36; // Left Control. Think Reference says 0x3B which is wrong
			}
			if(r_is_down != r_was_down) {
				return 0x3A; // Right Control -> option
			}
			return 0x36; // Left Control.
			// return(-1);

		// Command
		case VK_MENU: return 0x37; 

		// option
		case VK_LWIN: return 0x3a;
		case VK_RWIN: return 0x3a;



		case VK_F1: return 0x7a;
		case VK_F2: return 0x78;
		case VK_F3: return 0x63;
		case VK_F4: return 0x76;
		case VK_F5: return 0x60;
		case VK_F6: return 0x61;
		case VK_F7: return 0x62;
		case VK_F8: return 0x64;
		case VK_F9: return 0x65;
		case VK_F10: return 0x6d;
		case VK_F11: return 0x67;
		case VK_F12: return 0x6f;

		case VK_PRINT: return 0x69;
		case VK_SNAPSHOT: return 0x69;

		case VK_PAUSE: return 0x7f;  // Power key

		case VK_NUMPAD0: return 0x52;
		case VK_NUMPAD1: return 0x53;
		case VK_NUMPAD2: return 0x54;
		case VK_NUMPAD3: return 0x55;
		case VK_NUMPAD4: return 0x56;
		case VK_NUMPAD5: return 0x57;
		case VK_NUMPAD6: return 0x58;
		case VK_NUMPAD7: return 0x59;
		case VK_NUMPAD8: return 0x5b;
		case VK_NUMPAD9: return 0x5c;

		case VK_DECIMAL: return 0x41;
		case VK_ADD: return 0x45;
		case VK_SUBTRACT: return 0x4e;
		case VK_MULTIPLY: return 0x43;
		case VK_DIVIDE: return 0x4b;
		case VK_SEPARATOR: return 0x4c;



		// Toggle keys

		// Caps lock
		case VK_CAPITAL: 
			if(my_get_key_state(VK_CAPITAL) & TOGGLE_MASK) {
				if(down) {
					return 0x39;
				} else {
					// something wrong, -1 will make it in synch again
					return -1;
				}
			} else {
				if(down) {
					// something wrong, -1 will make it in synch again
					return -1;
				} else {
					return 0x39;
				}
			}
			// not reached
			break;

		// num lock, needs some work
		case 0x90: 
			return 0x47; // clear

		// scroll lock, come up with something here some day
		case 0x91: 
			/* MS DEV has this to say:
				Because the SetKeyboardState function alters the input state of the calling 
				thread and not the global input state of the system, an application cannot 
				use SetKeyboardState to set the num lock, caps lock, or scroll lock indicator 
				lights on the keyboard. 
			*/

			/*
			BYTE state[256];
			GetKeyboardState(state);
			if( state[0x91] & TOGGLE_MASK ) {
				state[0x91] &= ~TOGGLE_MASK;
				SetKeyboardState(state);
			}
			*/
			return 0x6b; // F14, better than nothing

	}
	return(-1);
}

static int do_key_down(HWND hWnd, UINT uMessage, WPARAM wparam, LPARAM lparam, int alt )
{
	int key;

	if((lparam & 0x40000000) == 0) {
		key = key_win2mac(wparam,true);

		/*
		char s[100];
		wsprintf( s, "Key: 0x%X -> 0x%X\r\n", (int)wparam, (int)key );
		OutputDebugString(s);
		*/

		if(key >= 0) {
			// TODO: use interlocked functions
			if(!mac_keys_down[key]) {
				ADBKeyDown((int)key);
				mac_keys_down[key] = 1;
			}
			return(1);
		}
	}
	return(1);
}

static int do_key_up( HWND hWnd, UINT uMessage, WPARAM wparam, LPARAM lparam, int alt )
{
	int key;
	key = key_win2mac(wparam,false);
	if(key >= 0) {
		// TODO: use interlocked functions
		if(mac_keys_down[key]) {
			ADBKeyUp(key);
			mac_keys_down[key] = 0;
		}
		return(1);
	}
	return(0);
}

static void suspend_emulation(void)
{
	if(CPUThread_handle && hMainWnd) {
		char caption[80];
		wsprintf( caption, "%s [slow mode]", GetString(STR_WINDOW_TITLE) );
		SetWindowText( hMainWnd, caption );
		SetThreadPriority( CPUThread_handle, THREAD_PRIORITY_BELOW_NORMAL );
	}
}

static void resume_emulation(void)
{
	if(CPUThread_handle && hMainWnd) {
		SetThreadPriority( CPUThread_handle, THREAD_PRIORITY_NORMAL );
		SetWindowText( hMainWnd, GetString(STR_WINDOW_TITLE) );
	}
}

static void update_floppy_keys(void)
{
	SHORT l = my_get_key_state(VK_LCONTROL);
	SHORT r = my_get_key_state(VK_RCONTROL);

	is_floppy_keys_down = false;

	if( (l & DOWN_MASK) != 0 && (r & DOWN_MASK) != 0 ) {
		if(floppy_keys_ok_time) {
			if(GetTickCount() >= floppy_keys_ok_time) {
				is_floppy_keys_down = true;
			}
		} else {
			floppy_keys_ok_time = GetTickCount() + FLOPPY_KEYS_SECONDS*1000;
		}
	} else {
		floppy_keys_ok_time = 0;
	}
}

bool is_main_floppy_reload(void)
{
	return(is_floppy_keys_down);
}

int get_floppy_detection_seconds(void)
{
	return(FLOPPY_KEYS_SECONDS);
}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	int answer;

	switch (msg) {
		case WM_CLOSE:
			ShowCursor(TRUE);
			answer = MessageBox(
				hWnd, 
				"You should normally close Basilisk II only from \"Special\"/\"Shut Down\".\r\n"
				"Really close now (some data may be lost)?", 
				GetString(STR_WINDOW_TITLE), 
				MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2
			);
			ShowCursor(FALSE);
			if(answer != IDNO) {
				DestroyWindow(hWnd);
			}
			break;
		case WM_KILLFOCUS:
			int i;
			for( i=0; i<MAX_KEYS; i++ ) {
				if(mac_keys_down[i]) {
					mac_keys_down[i] = 0;
					ADBKeyUp(i);
				}
			}
			// Hold down shift to keep it running at full speed
			if((GetAsyncKeyState(VK_SHIFT) & 0x8000) == 0) {
				suspend_emulation();
			}
			break;

		case WM_SETFOCUS:
			resume_emulation();
			break;

		case WM_CHAR:
		case WM_DEADCHAR:
		case WM_SYSDEADCHAR:
			break;

		case WM_PALETTECHANGED:
			D(bug("WM_PALETTECHANGED\r\n"));
			if (hWnd != (HWND) wParam) {
				D(bug("Not our window, updating palette\r\n"));
				update_palette();
				return DefWindowProc(hWnd, msg, wParam, lParam);
			}
			break;
			
		case WM_QUERYNEWPALETTE:
			D(bug("WM_QUERYNEWPALETTE\r\n"));
			return(update_palette());
			break;

		case WM_PAINT:
			Screen_Draw_All( hWnd );
			return 0;
			break;

		case WM_ERASEBKGND:
			// This is needed to prevent flashing
			break;

		case WM_CREATE:
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		case WM_MOVE:
			break;

    case WM_SIZE:
			// Not really used now

			if(wParam == SIZE_MINIMIZED) {
				// Hold down shift to keep it running at full speed
				if((GetAsyncKeyState(VK_SHIFT) & 0x8000) == 0) {
					suspend_emulation();
				}
			} else if(wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED) {
				resume_emulation();
			}
			break;

		case WM_KEYDOWN:
			update_floppy_keys();
			if(do_key_down(hWnd,msg,wParam,lParam,0)==0) return(0);
			else return DefWindowProc(hWnd, msg, wParam, lParam);
			break;

		case WM_KEYUP:
			update_floppy_keys();
			if(do_key_up(hWnd,msg,wParam,lParam,0)==0) return(0);
			else return DefWindowProc(hWnd, msg, wParam, lParam);
			break;

		case WM_SYSKEYDOWN:
			if(do_key_down(hWnd,msg,wParam,lParam,1)==0) return(0);
			else return DefWindowProc(hWnd, msg, wParam, lParam);
			break;

		case WM_SYSKEYUP:
			if(do_key_up(hWnd,msg,wParam,lParam,1)==0) return(0);
			else return DefWindowProc(hWnd, msg, wParam, lParam);
			break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			// This is atomic
			ADBMouseDown(0);
			break;

		case WM_LBUTTONUP:
			// This is atomic
			ADBMouseUp(0);
			if(capturing) {
				capturing = false;
				ReleaseCapture();
			}
			break;

		case WM_MOUSEMOVE:
			POINT mp;

			// Better to query the position again. There may be several mouse move
			// messages in the queue, and we want to update the last position
			// as soon as possible. It is better to lose one message than show
			// a "rubber-string" effect which can drive anybody nuts

			// mp.x = (int) LOWORD(lParam);
			// mp.y = (int) HIWORD(lParam);
			GetCursorPos(&mp);

			ADBMouseMoved(mp.x,mp.y);
			if(!capturing) {
				if( wParam & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON) ) {
					capturing = true;
					SetCapture(hWnd);
				}
			}

		default:
			return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 1;
}

void QuitEmulator(void)
{
	// Save XPRAM
	XPRAMExit();

	// Exit video
	VideoExit();

	// Exit clipboard
	ClipExit();

	// Exit Time Manager
	TimerExit();

	// Exit serial ports
	SerialExit();

	// Exit network
	EtherExit();

	// Exit drivers
	SCSIExit();
	CDROMExit();
	DiskExit();
	SonyExit();

	// Delete ROM area
	if(ROMBaseHost) delete[] ROMBaseHost;

	// Delete RAM area
	if(RAMBaseHost) delete[] RAMBaseHost;

	// Exit system routines
	SysExit();

	// Exit preferences
	PrefsExit();

	/*
	if(intflag_mutex) {
		CloseHandle(intflag_mutex);
		intflag_mutex = 0;
	}
	*/

	exit(0);
}

static void initialize(void)
{
	OSVERSIONINFO osv;

	win_os = VER_PLATFORM_WIN32_WINDOWS;
	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if(GetVersionEx( &osv )) {
		if(osv.dwPlatformId == VER_PLATFORM_WIN32s) {
			ErrorAlert("This program does not run on on Win32s.");
			QuitEmulator();
		} else {
			win_os = osv.dwPlatformId;
		}
	}
	
	timer_init();

	/*
	intflag_mutex = CreateMutex( 0, FALSE, "Basilisk II intflag mutex" );
	if(intflag_mutex == 0) {
		ErrorAlert("Cannot create Basilisk II intflag mutex.");
		QuitEmulator();
	}
	*/

	// Initialize variables
	RAMBaseHost = NULL;
	ROMBaseHost = NULL;

	srand(time(NULL));
	tzset();

	// Print some info
	// printf(GetString(STR_ABOUT_TEXT1), VERSION_MAJOR, VERSION_MINOR);
	// printf(" %s\n", GetString(STR_ABOUT_TEXT2));

	// Read preferences
	PrefsInit();

	// Init system routines
	SysInit();

	// Not used?
	CPUIs68060 = false;

	// Set (fake) CPU and FPU type (UAE emulation)
	// CPU type (2 = 68020, 3 = 68030, 4 = 68040/060)
	CPUType = 2;

	// FPU type (0 = no FPU, 1 = 68881, 2 = 68882)
	FPUType = 1;
	if (!PrefsFindBool("nofpu")) FPUType = 0;

#define ONEMB (1024*1024)
#define FOURMB (1024*1024)

	// Create area for Mac RAM
	RAMSize = PrefsFindInt32("ramsize");
	if (RAMSize < ONEMB) {
		WarningAlert(GetString(STR_SMALL_RAM_WARN));
		RAMSize = ONEMB;
	}

	// Round up. Must be aligned.
	if(RAMSize % FOURMB) {
		RAMSize = (RAMSize/FOURMB + 1) * FOURMB;
	}

	RAMBaseHost = new uint8[RAMSize];
	if(!RAMBaseHost) {
		ErrorAlert(GetString(STR_NO_RAM_AREA_ERR));
		QuitEmulator();
	}

#if REAL_ADDRESSING
	RAMBaseMac = (uint32)RAMBaseHost;
#else
	RAMBaseMac = 0;
#endif

	RAMBaseDiff = (int32)RAMBaseHost - (int32)RAMBaseMac;

	// Create area for Mac ROM
	ROMBaseHost = new uint8[0x100000];
	if(!ROMBaseHost) {
		ErrorAlert(GetString(STR_NO_ROM_AREA_ERR));
		QuitEmulator();
	}
#if REAL_ADDRESSING
	ROMBaseMac = (uint32)ROMBaseHost;
#else
	ROMBaseMac = 0x40800000;
#endif

	ROMBaseDiff = (int32)ROMBaseHost - (int32)ROMBaseMac;


	// Load Mac ROM
	HFILE rom_fd = _lopen(ROM_FILE_NAME, OF_READ);
	if (rom_fd < 0) {
		ErrorAlert(GetString(STR_NO_ROM_FILE_ERR));
		QuitEmulator();
	}
	ROMSize = _llseek(rom_fd, 0, SEEK_END);
	if (ROMSize != 1024*1024 && ROMSize != 512*1024) {
		ErrorAlert(GetString(STR_ROM_SIZE_ERR));
		_lclose(rom_fd);
		QuitEmulator();
	}
	_llseek(rom_fd, 0, SEEK_SET);
	if (_lread(rom_fd, ROMBaseHost, ROMSize) != ROMSize) {
		ErrorAlert(GetString(STR_ROM_FILE_READ_ERR));
		_lclose(rom_fd);
		QuitEmulator();
	}
	_lclose(rom_fd);

	// Load XPRAM
	XPRAMInit();

	// Set boot volume
	int16 i16 = PrefsFindInt16("bootdrive");
	XPRAM[0x78] = i16 >> 8;
	XPRAM[0x79] = i16 & 0xff;

	i16 = PrefsFindInt16("bootdriver");
	XPRAM[0x7a] = i16 >> 8;
	XPRAM[0x7b] = i16 & 0xff;

	// Init drivers

	// TODO: device access should be protected by mutexes or critical sections

	SonyInit();
	DiskInit();
	CDROMInit();
	SCSIInit();

	// Init serial ports
	SerialInit();

	// Init network
	EtherInit();

	// Init Time Manager
	TimerInit();

	// Init clipboard
	ClipInit();

#if !REAL_ADDRESSING
	// Initialize UAE memory banks
	memory_init();
#endif

	// Init video
	if (!VideoInit()) {
		ErrorAlert("Failed to init video.");
		QuitEmulator();
	}

#if !REAL_ADDRESSING
	// Reinitialize UAE memory banks (VideoInit() set the frame buffer)
	memory_init();
#endif

	// Install ROM patches
	if (!PatchROM()) {
		ErrorAlert(GetString(STR_UNSUPPORTED_ROM_TYPE_ERR));
		QuitEmulator();
	}

	DWORD tid; // Temp. Don't need thread ids.

	// Start 60Hz thread
	CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)tick_func, (LPVOID) hMainWnd, 0, &tid);

	// Start 68k and jump to ROM boot routine
	CPUThread_handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)cpu_thread, (LPVOID) hMainWnd, 0, &tid);

	// Start XPRAM watchdog thread
	CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)xpram_func, (LPVOID)0, 0, &tid);

	// Boost GUI operations to one notch above interrupt thread.
	// Must choose which one is more important, insanely accurate interrupts
	// or smooth mouse movement. I feel that mouse is the more important

	// Consequently, palette animation is now executed at top priority
	// which is not what I really wanted. Could be a problem in 3d shooters
	// when you get hit
	SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_HIGHEST );
}


void WarningAlert(const char *text)
{
	ShowCursor(TRUE);
	MessageBox(hMainWnd,text,GetString(STR_WINDOW_TITLE),MB_OK|MB_ICONINFORMATION);
	ShowCursor(FALSE);
}

void ErrorAlert(const char *text)
{
	ShowCursor(TRUE);
	MessageBox(hMainWnd,text,GetString(STR_WINDOW_TITLE),MB_OK|MB_ICONSTOP);
	ShowCursor(FALSE);
}

void __inline__ SetInterruptFlag(uint32 flag)
{
	// Too big performance hit doing this properly.
	// We MAY lose sometimes mouse or serial interrupt,
	// that's the price to pay. Arguably acceptable.
	// WaitForSingleObject(intflag_mutex,INFINITE);
	InterruptFlags |= flag;
	// ReleaseMutex(intflag_mutex);
}

void ClearInterruptFlag(uint32 flag)
{
	// WaitForSingleObject(intflag_mutex,INFINITE);
	InterruptFlags &= ~flag;
	// ReleaseMutex(intflag_mutex);
}

/*
 *  Code was patched, flush caches if neccessary (i.e. when using a real 680x0
 *  or a dynamically recompiling emulator)
 */

#if EMULATED_68K
void FlushCodeCache(void *start, uint32 size)
{
}
#endif

DWORD WINAPI cpu_thread(LPVOID param)
{
	init_m68k();
	m68k_reset();
	m68k_go(true);
	return 0;
}

/*
 *  60Hz thread
 */

#if DEBUG
static int counter_one_sec = 0;
static int counter_60_hz = 0;
#endif




#if(_WIN32_WINNT >= 0x0400)

enum { 
	TIMER_60_HZ=0,
	TIMER_ONE_SEC,
	TIMER_COUNT // last marker
};

// The NT timer version is very accurate,
// but resolution is 1ms. Thus ve have a solid 62.5 Hz timer.
// We cannot use this code however for performance reasons.

DWORD WINAPI tick_func_winnt(LPVOID param)
{
	DWORD mac_boot_secs = 0;
	time_t t;
	HANDLE timers[TIMER_COUNT];
	LARGE_INTEGER trigger = {0,0}; // SetWaitableTimer() does not accept null
	int i;

	time( &t );
	t = t - timezone + (daylight ? 3600 : 0) + TIME_OFFSET;
	mac_boot_secs = t - GetTickCount()/1000 ;
	WriteMacInt32(0x20c, t);

	timers[TIMER_60_HZ] = CreateWaitableTimer( 0, 0, 0 );
	SetWaitableTimer( timers[TIMER_60_HZ], &trigger, 16, 0, 0, FALSE );

	timers[TIMER_ONE_SEC] = CreateWaitableTimer( 0, 0, 0 );
	SetWaitableTimer( timers[TIMER_ONE_SEC], &trigger, 1000, 0, 0, FALSE );

	for (;;) {
		i = WaitForMultipleObjects( 2, timers, FALSE, INFINITE );
		if( i == WAIT_OBJECT_0 ) {
			// Trigger 60Hz interrupt
#if DEBUG
			counter_60_hz++;
#endif
			SetInterruptFlag(INTFLAG_60HZ);
			TriggerInterrupt();
		} else if( i == WAIT_OBJECT_0+1 ) {
			// Pseudo Mac 1Hz interrupt, update local time
			D(bug("One sec timer %d,%d\r\n",++counter_one_sec,counter_60_hz));
			uint32 secs = mac_boot_secs + GetTickCount()/1000;
			WriteMacInt32(0x20c, secs);
		}
	}

	for( i=0; i<TIMER_COUNT; i++ ) {
		CancelWaitableTimer(timers[i]);
		CloseHandle(timers[i]);
	}

	return 0;
}
#endif // (_WIN32_WINNT >= 0x0400)




// This is an accurate timer implemented using two differential timers.
// Performance penalty is acceptable.

DWORD WINAPI tick_func_win9x(LPVOID param)
{
	int i, tick_counter = 0;
	DWORD mac_boot_secs = 0;
	time_t t;
	int32 ttable[60];

	// These may go negative so we use signed values, 
	// we still have about 414 days time frame
  int32 sleep_amount, should_have_ticks, should_have_ticks2;

	// One time code, may use doubles.
	// First setup a time target table. Split the time evenly.
	// The important thing is that the last entry is 1000
	for( i=0; i<60; i++ ) {
		ttable[i] = (int32)( 1000.0 * (i+1) / 60.0 + 0.5 );
	}
	// ... so we make it sure:
	ttable[59] = 1000;

	// Then convert the table to be a differential table.
	// ttable[0] is already ok.
	for( i=59; i>0; i-- ) { 
		ttable[i] -= ttable[i-1];
	}

#if DEBUG
	// And verify it.
	int checksum = 0;
	for( i=0; i<60; i++ ) checksum += ttable[i];
	D(bug("Timer table sum=%d\r\n",checksum));
#endif

	time( &t );

	t = t - timezone + (daylight ? 3600 : 0) + TIME_OFFSET;
	mac_boot_secs = t - GetTickCount()/1000 ;
	WriteMacInt32(0x20c, t);

	// This is not of too much use but does some good in reducing latency.
	// I would like to speed up the code that actually executes the interrupt.
	SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL );

	int ttinx = 0;

	// first value is an educated guess. will correct self later
	int estimated_rounds_left = 60; 

	should_have_ticks = GetTickCount();
	should_have_ticks2 = GetTickCount() + 1000;

	for (;;) {
		// simple self adjusting clock.
		if(++ttinx >= 60) ttinx = 0;
		should_have_ticks += ttable[ttinx];
		sleep_amount = should_have_ticks - (int)GetTickCount() - 1;
		if(sleep_amount >= 5) Sleep(sleep_amount);

		// Pseudo Mac 1Hz interrupt, update local time
		if (++tick_counter >= estimated_rounds_left) {
			D(bug("One sec timer %d,%d\r\n",++counter_one_sec,counter_60_hz));
			int32 now = GetTickCount();
			int32 secs = mac_boot_secs + now/1000;
			WriteMacInt32(0x20c, secs);

			should_have_ticks2 += 1000;
			// Note that it does not matter how we round the value below.
			// It will always correct itself.
			estimated_rounds_left = (should_have_ticks2 - now) / 60 + 1;
			tick_counter = 0;
		}

#if DEBUG
		counter_60_hz++;
#endif
		// Trigger 60Hz interrupt
		SetInterruptFlag(INTFLAG_60HZ);
		TriggerInterrupt();
	}
	return 0;
}

DWORD WINAPI tick_func(LPVOID param)
{
	if(win_os == VER_PLATFORM_WIN32_NT) {
#if (_WIN32_WINNT >= 0x0400)
		// This is *theoritically* the preferred way to go.
		// However, the performance hit is too big to be useable.

		//  return(tick_func_winnt(param));
		return(tick_func_win9x(param));

#else
		return(tick_func_win9x(param));
#endif
	} else {
		return(tick_func_win9x(param));
	}
}

// dummy
#define IDM_ABOUT 998

static void build_menu( HWND hWnd )
{
	char about[100];
	wsprintf( about, GetString(STR_ABOUT_TEXT1), VERSION_MAJOR, VERSION_MINOR );
	wsprintf( &about[strlen(about)], " %s", GetString(STR_ABOUT_TEXT2) );
	HMENU system_menu = GetSystemMenu( hWnd, FALSE );
	ChangeMenu( system_menu, 0, NULL, 999, MF_APPEND | MF_SEPARATOR );
  ChangeMenu( system_menu, 0, about, IDM_ABOUT, MF_APPEND | MF_STRING | MF_DISABLED );
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd )
{
	WNDCLASSEX wc;
	MSG msg;

	hInst = hInstance;

	memset( &wc, 0, sizeof(wc) );
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)GetStockObject(GRAY_BRUSH);
	wc.hInstance = hInstance;
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = GetString(STR_WINDOW_TITLE);
	wc.style = CS_DBLCLKS|CS_HREDRAW|CS_VREDRAW|CS_BYTEALIGNCLIENT;
	RegisterClassEx(&wc);

	hMainWnd = CreateWindowEx(
		WS_EX_APPWINDOW,
		GetString(STR_WINDOW_TITLE),
		GetString(STR_WINDOW_TITLE),
		WS_POPUP,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		HWND_DESKTOP,
		NULL,
		hInstance,
		NULL
	);
	if(!hMainWnd) {
		ErrorAlert("Failed to create main window.");
		return(0);
	}

	ShowCursor(FALSE);
	ShowWindow(hMainWnd, SW_MAXIMIZE);
	initialize();

	while ( GetMessage(&msg, NULL, 0, 0) ) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return(msg.wParam);
}
