/*
 *  sys_windows.cpp - System dependent routines for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include "sysdeps.h"
#include "main.h"
#include "prefs.h"
#include "user_strings.h"
#include "sys.h"
#include "errno.h"
#include "cdenable\ntcd.h"
#include "cdenable\cache.h"
#include "cdenable\eject_nt.h"
#include "winioctl.h"
#include "main_windows.h"

#include "debug.h"

#define MAX_DEVICE_NAME _MAX_PATH

// File handles are pointers to these structures
struct file_handle {
	HANDLE h;
	bool is_file;			// Flag: plain file or /dev/something?
	bool is_floppy;		// Flag: floppy device
	bool is_cdrom;		// Flag: CD-ROM device
	bool is_hd;				// Flag: Hard disk.
	bool read_only;		// Copy of Sys_open() flag
	off_t start_byte;	// Size of file header (if any)
	cachetype cache;
	char name[MAX_DEVICE_NAME];
	bool is_floppy_present;
};


static bool is_nt = false;

#define FLOPPY_ALIGN_MEMORY_SIZE 512
#define HD_ALIGN_MEMORY_SIZE 512
#define CD_READ_AHEAD_SECTORS 32

static char *sector_buffer = 0;

static void check_os(void)
{
	OSVERSIONINFO osv;

	is_nt = false;

	osv.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if(GetVersionEx( &osv )) {
		switch(osv.dwPlatformId) {
			case VER_PLATFORM_WIN32s:
				ErrorAlert("Cannot run on Win32s");
				QuitEmulator();
				// not reached
				break;
			case VER_PLATFORM_WIN32_NT:
				is_nt = true;
				break;
			default:
				is_nt = false;
		}
	}
}

static int cd_read( 
	HANDLE hcd, 
	cachetype *cptr,
	ULONG LBA, 
	int count, 
	char *buf
)
{
	ULONG l1, l2, cc;
	int i, c_count, got_bytes, nblocks, s_inx, ss, first_block;
	int ok_bytes = 0;
	char *ptr, *ttptr = 0, *tmpbuf;

	if(hcd == 0 || count <= 0) return(0);

	ss = 2048;
	l1 = (LBA / ss) * ss;
	l2 = ((LBA+count-1+ss)/ss) * ss;
	cc = l2-l1;
	nblocks = cc / ss;
	first_block = LBA / ss;

	ptr = buf;
	s_inx = LBA-l1;
	c_count = ss - s_inx;
	if(c_count > count) c_count = count;
	for(i=0; i<nblocks; i++) {
		if(cache_get( cptr, first_block+i, sector_buffer )) {
			memcpy( ptr, sector_buffer+s_inx, c_count );
			ok_bytes += c_count;
			ptr += c_count;
			s_inx = 0;
			c_count = ss;
			if(c_count > count-ok_bytes) c_count = count-ok_bytes;
		} else {
			break;
		}
	}

	if(i != nblocks && count != ok_bytes) {
		int bytes_left, blocks_left, alignedleft;

		bytes_left = count - ok_bytes;
		blocks_left = nblocks - i;

		// NEW read ahead code:
		int ahead = CD_READ_AHEAD_SECTORS;
		if(blocks_left < ahead) {
			nblocks += (ahead - blocks_left);
			blocks_left = ahead;
		}

		alignedleft = blocks_left*ss;

		tmpbuf = (char *)VirtualAlloc( 
				NULL, alignedleft, 
				MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(tmpbuf) {
			if(is_nt) {
				got_bytes = CdenableSysReadCdBytes( hcd, (first_block+i)*ss, alignedleft, tmpbuf );
			}

			if(got_bytes != alignedleft) {
				// should never happen
				// Yes it does ...
				if(got_bytes < 0) got_bytes = 0;
				if(c_count > got_bytes) c_count = got_bytes;
				if(c_count > 0) {
					ttptr = tmpbuf;
					memcpy( ptr, ttptr+s_inx, c_count );
					ok_bytes += c_count;
				}
				VirtualFree( tmpbuf, 0, MEM_RELEASE  ); 
				return(ok_bytes); 
			}
			ttptr = tmpbuf;
			for( ; i<nblocks; i++ ) {
				if(c_count > 0) {
					memcpy( ptr, ttptr+s_inx, c_count );
					ok_bytes += c_count;
					ptr += c_count;
				}
				s_inx = 0;
				c_count = ss;
				if(c_count > count-ok_bytes) c_count = count-ok_bytes;
				cache_put( cptr, first_block+i, ttptr, ss );
				ttptr += ss;
			}
			VirtualFree( tmpbuf, 0, MEM_RELEASE  ); 
		}
	}

	return(ok_bytes);
}

size_t floppy_read( HANDLE hfloppy, off_t LBA, size_t count, char *buf )
{
	DWORD bytes_read;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!hfloppy) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (FLOPPY_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
	} else {
		rd_buffer = buf;
	}

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

  if (!ReadFile (hfloppy, rd_buffer, count, &bytes_read, &overl)) {
		bytes_read = count = 0;
	}
	if(aux_buffer) {
		if(count) memcpy( buf, rd_buffer, count );
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_read);
}

size_t nt_floppy_write( HANDLE hfloppy, ULONG LBA, int count, char *buf )
{
	DWORD bytes_written;
	OVERLAPPED overl;
	char *aux_buffer = 0;

	if(!hfloppy) return(0);
	if(count == 0) return(0);

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (FLOPPY_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		memcpy( aux_buffer, buf, count );
		buf = aux_buffer;
	}
	
  if (WriteFile (hfloppy, buf, count, &bytes_written, &overl) &&
			(int)bytes_written == count)
	{
		;
	} else {
		bytes_written = 0;
	}
	if(aux_buffer) VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	return(bytes_written);
}

/*
 *  Initialization
 */

void SysInit(void)
{
	check_os();

	sector_buffer = (char *)VirtualAlloc( NULL, 8192, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );

 	if(is_nt) {
		CdenableSysInstallStart();
	}
}

/*
 *  Deinitialization
 */

void SysExit(void)
{
	if(sector_buffer) {
		VirtualFree( sector_buffer, 0, MEM_RELEASE  ); 
		sector_buffer = 0;
	}
}

static void scan_floppies( char *type )
{
	char rootdir[20], letter;
	int i;

	for( letter = 'A'; letter <= 'B'; letter++ ) {
		i = (int)( letter - 'A' );
		wsprintf( rootdir, "%c:\\", letter );
		if(GetDriveType( rootdir ) == DRIVE_REMOVABLE) {
			PrefsAddString(type, rootdir);
		}
	}
}

/*
 *  This gets called when no "floppy" prefs items are found
 *  It scans for available floppy drives and adds appropriate prefs items
 */

void SysAddFloppyPrefs(void)
{
	scan_floppies("floppy");
}

/*
 *  This gets called when no "disk" prefs items are found
 *  It scans for available HFS volumes and adds appropriate prefs items
 */

void enum_hard_files( char *dir, char *extension )
{
	char mask[_MAX_PATH], path[_MAX_PATH];
	HANDLE fh;
	WIN32_FIND_DATA FindFileData;
	int ok;

	wsprintf( mask, "%s*.%s", dir, extension );
	fh = FindFirstFile( mask, &FindFileData );
	ok = fh != INVALID_HANDLE_VALUE;
	while(ok) {
		ok = FindNextFile( fh, &FindFileData );
		sprintf( path, "%s%s", dir, FindFileData.cFileName );
		PrefsAddString("disk", path);
	}
	if(fh != INVALID_HANDLE_VALUE) FindClose( fh );
}

void enum_hd_partitions(void)
{
	char rootdir[20], letter;
	int i;
	UINT type;

	for( letter = 'C'; letter <= 'Z'; letter++ ) {
		i = (int)( letter - 'A' );
		wsprintf( rootdir, "%c:\\", letter );
		type = GetDriveType( rootdir );
		if(type == DRIVE_FIXED) {
			PrefsAddString("disk", rootdir);
		}
	}
}

void SysAddDiskPrefs(void)
{
	char dir[_MAX_PATH];

	GetCurrentDirectory(sizeof(dir),dir);
	if(strlen(dir) > 0 && dir[strlen(dir)-1] != '\\') strcat(dir,"\\");

	enum_hard_files( dir, "hfv" );
	enum_hard_files( dir, "dsk" );
	// scan_floppies("disk");
	enum_hd_partitions();
}


/*
 *  This gets called when no "cdrom" prefs items are found
 *  It scans for available CD-ROM drives and adds appropriate prefs items
 */

void SysAddCDROMPrefs(void)
{
	// Don't scan for drives if nocdrom option given
	if (!PrefsFindBool("nocdrom")) {
		char rootdir[20], letter;
		int i;

		for( letter = 'C'; letter <= 'Z'; letter++ ) {
			i = (int)( letter - 'A' );
			wsprintf( rootdir, "%c:\\", letter );
			if(GetDriveType( rootdir ) == DRIVE_CDROM) {
				PrefsAddString("cdrom", rootdir);
			}
		}
	}
}


/*
 *  Add default serial prefs (must be added, even if no ports present)
 */

void SysAddSerialPrefs(void)
{
	PrefsAddString("seriala", "COM1");
	PrefsAddString("serialb", "COM2");
}

static bool is_floppy_readable( file_handle *fh )
{
	UINT prevmode = SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);
	char dummy[0x200];
	bool result = false;

	if( fh && fh->h != INVALID_HANDLE_VALUE ) {
		if(floppy_read( fh->h, 0, sizeof(dummy), dummy ) == sizeof(dummy)) {
			result = true;
		}
  }
	SetErrorMode(prevmode);
	return( result );
}

/*
 *  Open file/device, create new file handle (returns NULL on error)
 */

void *Sys_open(const char *_name, bool read_only)
{
	HANDLE h;
	off_t size;
	file_handle *fh = 0;

	char dnamebuf[64];
	char dname[64];
	char dnamel[64];
	char *name, local_name[_MAX_PATH];

	bool physical = false; // open physical/logical volume

	strcpy( local_name, _name );
	strupr( local_name );

	name = local_name;

	if(strncmp(name,"PHYSICAL ",9) == 0) {
		physical = true;
		name += 9;
		while(*name == ' ') name++;
	}

	// Normalize floppy / cd path.
	// User may give "disk A:" or "disk A:\" or "disk A" or "disk physical A:"

	int name_len = strlen(name);

	if(name_len == 1 && isalpha(*name)) {
		strcat( name, ":\\" );
	}

	if(name_len > 0 && name[name_len-1] == ':') {
		strcat( name, "\\" );
	}

	name_len = strlen(name);

	D(bug("Sys_open(%s, %s)\n", name, read_only ? "read-only" : "read/write"));

	if(name_len > 0 && name[name_len-1] == '\\') {
		UINT type = GetDriveType(name);
		if(type == DRIVE_FIXED) {
			// !!!!!! change this at your own risk.
			read_only = true;

			if(is_nt) {
				*dname = 0;
				if(physical) {
					sprintf(dname,"%c:",(char)toupper(*name));
					if ( QueryDosDevice(dname, dnamebuf, sizeof(dnamebuf)) == 0) {
						return(0);
					} else {
						char *h, *p;
						h = strstr( dnamebuf, "\\Harddisk" );
						p = strstr( dnamebuf, "\\Partition1" );
						if(h == 0 || p == 0) return(0);
						h += 9;
						int inx = atoi(h);
						sprintf( dnamebuf, "\\\\.\\PHYSICALDRIVE%d", inx );
					}
				} else {
					sprintf( dnamebuf, "\\\\.\\%c:", (char)toupper(*name) );
				}

				h = CreateFile( 
					dnamebuf, 
					read_only ? GENERIC_READ : GENERIC_READ|GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE,
					0, 
					OPEN_EXISTING, 
					FILE_FLAG_NO_BUFFERING,
					0 
				);
				if( read_only == 0 && h == INVALID_HANDLE_VALUE ) {
					read_only = true;
					h = CreateFile( 
						dnamebuf, 
						GENERIC_READ, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						0, 
						OPEN_EXISTING, 
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED,
						0 
					);
					if( h != INVALID_HANDLE_VALUE ) {
						char msg[200];
						wsprintf( 
							msg, 
							"Could not get read/write access to the drive \"%s\", mounting read-only.",
							fh->name
						);
						WarningAlert( msg );
					}
				}
				
				if( read_only == 0 && h != INVALID_HANDLE_VALUE ) {
					DWORD dwBytesReturned = 0;
					if(!DeviceIoControl(h,
						FSCTL_LOCK_VOLUME,
						NULL, 0,
						NULL, 0,
						&dwBytesReturned,
						NULL))
					{
						CloseHandle(h);
						h = INVALID_HANDLE_VALUE;
					}
				}

				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					fh->h = h;
					fh->is_file = false;
					fh->is_floppy = false;
					fh->is_cdrom = false;
					fh->is_hd = true;
					fh->read_only = read_only;
					fh->start_byte = 0;
					strcpy( fh->name, name );
				}
			}
		} else if(GetDriveType(name) == DRIVE_REMOVABLE) {
			if(is_nt) {
				int rc;
				sprintf(dname,"Basilisk2_floppy_%c",(char)toupper(*name));

				// TODO: fails on systems where \Device\Floppy0 is B:
				sprintf(dnamel,"\\Device\\Floppy%d",(int)toupper(*name) - 'A');

				if ( QueryDosDevice(dname, dnamebuf, sizeof(dnamebuf)) == 0) {
					rc = GetLastError();
					if ( rc == ERROR_FILE_NOT_FOUND ) {
						if ( DefineDosDevice(DDD_RAW_TARGET_PATH, dname, dnamel) == 0 ) {
							return( 0);
						}
					} else {
						return(0);
					}
				}
				sprintf(dnamebuf,"\\\\.\\%s", dname);

				UINT prevmode = SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);

				sprintf(dname,"\\\\.\\%c:",(char)*name);
				h = CreateFile (dname, GENERIC_READ|GENERIC_WRITE,
											 FILE_SHARE_READ,
											 NULL, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS,
											 0);
				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					fh->h = h;
					fh->is_file = false;
					fh->is_floppy = true;
					fh->is_cdrom = false;
					fh->is_hd = false;
					fh->read_only = read_only;
					fh->start_byte = 0;
					strcpy( fh->name, name );
					// memset( &fh->cache, 0, sizeof(cachetype) );
					// cache_init( &fh->cache );
					if (!PrefsFindBool("nofloppyboot")) {
						fh->is_floppy_present = is_floppy_readable(fh);
					} else {
						fh->is_floppy_present = false;
					}
  			}
				SetErrorMode(prevmode);
			}
		} else if(GetDriveType(name) == DRIVE_CDROM) {
			read_only = true;
			if(is_nt) {
				sprintf(dname,"\\\\.\\%c:",(char)*name);
				h = CreateFile (dname, GENERIC_READ,
											 FILE_SHARE_READ|FILE_SHARE_WRITE,
											 NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
											 NULL);
				if( h != INVALID_HANDLE_VALUE ) {
					fh = new file_handle;
					fh->h = h;
					fh->is_file = false;
					fh->is_floppy = false;
					fh->is_cdrom = true;
					fh->is_hd = false;
					fh->read_only = read_only;
					fh->start_byte = 0;
					strcpy( fh->name, name );
					memset( &fh->cache, 0, sizeof(cachetype) );
					cache_init( &fh->cache );
				}
			}
		} else if(GetDriveType(name) == DRIVE_REMOVABLE) {
		}
	} else {
		// hard file.
		// Check if write access is allowed, set read-only flag if not
		// Open file/device (for exclusive access, to protect data integrity)
		h = CreateFile( 
			name,
			GENERIC_READ | GENERIC_WRITE,
			0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL
		);

		if(h == INVALID_HANDLE_VALUE) {
			read_only = true;
			h = CreateFile( 
				name,
				GENERIC_READ,
				0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL
			);
		}

		if(h != INVALID_HANDLE_VALUE) {
			fh = new file_handle;
			fh->h = h;
			fh->is_file = true;
			fh->is_floppy = false;
			fh->is_cdrom = false;
			fh->is_hd = false;
			fh->read_only = read_only;
			fh->start_byte = 0;
			strcpy( fh->name, name );
			if (fh->is_file) {
				// Detect header length: if the file size is not a multiple of 512,
				// the remainder at the beginning is considered the header and skipped
				size = (long)GetFileSize(h,NULL);
				if (size == 838484)	// 800K DiskCopy image
					fh->start_byte = 84;
				else
					fh->start_byte = size % 512;
			}
		}
	}

	return fh;
}


/*
 *  Close file/device, delete file handle
 */

void Sys_close(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (fh) {
		if( fh->is_cdrom  ) {
			cache_final( &fh->cache );
		}
		if( !fh->read_only && fh->is_hd && fh->h ) {
			DWORD dwBytesReturned = 0;
			DeviceIoControl(
				fh->h,
				FSCTL_UNLOCK_VOLUME,
				NULL, 0,
				NULL, 0,
				&dwBytesReturned,
				NULL
			);
		}
		if( fh->h ) {
			CloseHandle( fh->h );
			fh->h = 0;
		}
		delete fh;
	}
}

size_t hd_read( HANDLE hcd, ULONG LBA, int count, char *buf )
{
	DWORD bytes_read;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!hcd) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (HD_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
	} else {
		rd_buffer = buf;
	}

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

  if (!ReadFile (hcd, rd_buffer, count, &bytes_read, &overl)) {
		bytes_read = count = 0;
	}
	if(aux_buffer) {
		if(bytes_read) memcpy( buf, rd_buffer, bytes_read );
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_read);
}

int nt_hd_write( HANDLE hcd, ULONG LBA, int count, char *buf )
{
	DWORD bytes_written;
	OVERLAPPED overl;
	char *aux_buffer = 0;
	char *rd_buffer;

	if(!hcd) return(0);

	if(count == 0) return(0);

	// If not aligned to sector boundary -> must use another buffer
	if( (unsigned long)buf & (HD_ALIGN_MEMORY_SIZE-1) ) {
		aux_buffer = (char *)VirtualAlloc( 
				NULL, count, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
		if(!aux_buffer) return(0);
		rd_buffer = aux_buffer;
		memcpy( rd_buffer, buf, count );
	} else {
		rd_buffer = buf;
	}

	overl.Offset = LBA;
	overl.OffsetHigh = 0;
	overl.hEvent = NULL;

  if (!WriteFile (hcd, rd_buffer, count, &bytes_written, &overl)) {
		if(GetLastError() == ERROR_WRITE_PROTECT) {
			// AfxMessageBox( "The media is write protected." );
		}
		bytes_written = count = 0;
	}

	if(aux_buffer) {
		VirtualFree( aux_buffer, 0, MEM_RELEASE  ); 
	}
	return(bytes_written);
}

/*
 *  Read "length" bytes from file/device, starting at "offset", to "buffer",
 *  returns number of bytes read (or 0)
 */

size_t Sys_read(void *arg, void *buffer, off_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	DWORD bytes_read = 0;

	if (!fh) return 0;

	if(fh->is_file) {
		if( SetFilePointer( fh->h, offset + fh->start_byte, NULL, FILE_BEGIN ) != 0xFFFFFFFF ) {
			if(!ReadFile( fh->h, buffer, length, &bytes_read, NULL)) {
				bytes_read = 0;
			}
		}
	} else if(fh->is_floppy || fh->is_cdrom || fh->is_hd) {
		size_t bytes_left, try_bytes, got_bytes;
		char *b = (char *)buffer;
		bytes_left = length;
		while(bytes_left) {
			try_bytes = min( bytes_left, 32768 );
			if(fh->is_cdrom) {
				got_bytes = cd_read( fh->h, &fh->cache, offset, try_bytes, b );
			} else if(fh->is_floppy) {
				got_bytes = floppy_read( fh->h, offset, length, (char*)buffer );
				if(got_bytes != length) {
					fh->is_floppy_present = is_floppy_readable(fh);
				}
			} else if(fh->is_hd) {
				got_bytes = hd_read( fh->h, offset, try_bytes, (char*)b );
			}
			b += got_bytes;
			offset += got_bytes;
			bytes_read += got_bytes;
			bytes_left -= got_bytes;
			if(got_bytes != try_bytes) {
				bytes_left = 0;
			}
		}
	}

	return(bytes_read);
}


/*
 *  Write "length" bytes from "buffer" to file/device, starting at "offset",
 *  returns number of bytes written (or 0)
 */

size_t Sys_write(void *arg, void *buffer, off_t offset, size_t length)
{
	file_handle *fh = (file_handle *)arg;
	DWORD bytes_written = 0;

	if(!fh) return 0;

	if(fh->is_file) {
		if( SetFilePointer( fh->h, offset + fh->start_byte, NULL, FILE_BEGIN ) != 0xFFFFFFFF ) {
			if(!WriteFile( fh->h, buffer, length, &bytes_written, NULL)) {
				bytes_written = 0;
			}
		}
	} else if(fh->is_hd) {
		if(fh->read_only) {
			OutputDebugString( "Attempt to write to HD partition rejected.\r\n" );
		} else {
			bytes_written = nt_hd_write( fh->h, offset, length, (char *)buffer );
		}
	} else if(fh->is_floppy) {
		bytes_written = nt_floppy_write( fh->h, offset, length, (char *)buffer );
		if(bytes_written != length) {
			fh->is_floppy_present = is_floppy_readable(fh);
		}
	}
	return(bytes_written);
}

/*
 *  Return size of file/device (minus header)
 */

off_t SysGetFileSize(void *arg)
{
	off_t length = 0;

	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	if(fh->h) {
		if (fh->is_file) {
			length = (long)GetFileSize(fh->h,NULL);
		} else if (fh->is_floppy) {
			// test. probably never called
			length = 80 * 2 * 18 * 512;
		} else if (fh->is_cdrom) {
			// test. probably never called
			length = 0x28A00000;
		} else if (fh->is_hd) {
			// test. probably never called
			length = 0;
		}
	}

	return length;
}


/*
 *  Eject volume (if applicable)
 */

void SysEject(void *arg)
{
	DWORD dwBytesReturned;
	file_handle *fh = (file_handle *)arg;

	if (!fh) return;

	if (fh->is_floppy) {
		char msg[300];
		if(fh->is_floppy_present) {
			fh->is_floppy_present = false;
			DeviceIoControl( 
				fh->h,
				IOCTL_STORAGE_EJECT_MEDIA,
				NULL, 0,
				NULL, 0,
				&dwBytesReturned,
				NULL
			);
			wsprintf( 
				msg, 
				"The floppy was ejected from drive \"%c:\".\r\n"
				"Hold both left and right control keys down for %d seconds to reload the floppy.", 
				(char)toupper(*fh->name),
				get_floppy_detection_seconds()
			);
			WarningAlert( msg );
		}
	} else if (fh->is_cdrom) {
		if(is_nt) {
			// EjectVolume(toupper(*fh->name),false);

			for(int i=0; i<10; i++) PreventRemovalOfVolume(fh->h,false);

			if (!PrefsFindBool("nocdrom")) {
				DeviceIoControl( 
					fh->h,
					IOCTL_STORAGE_EJECT_MEDIA,
					NULL, 0,
					NULL, 0,
					&dwBytesReturned,
					NULL
				);
			}
			cache_clear( &fh->cache );
		}
	}
}

/*
 *  Format volume (if applicable)
 */

bool SysFormat(void *arg)
{
	file_handle *fh = (file_handle *)arg;

	if (!fh) return false;

	if(fh->read_only) return false;

	return true;
}


/*
 *  Check if file/device is read-only (this includes the read-only flag on Sys_open())
 */

bool SysIsReadOnly(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	return fh->read_only;
}


/*
 *  Check if the given file handle refers to a fixed or a removable disk
 */

bool SysIsFixedDisk(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return true;

	if (fh->is_file || fh->is_hd)
		return true;
	else if (fh->is_floppy || fh->is_cdrom)
		return false;
	else
		return true;
}


/*
 *  Check if a disk is inserted in the drive (always true for files)
 */

bool SysIsDiskInserted(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	bool result = false;

	if (!fh) return false;

	if (fh->is_file || fh->is_hd) {
		result = true;
	} else if (fh->is_floppy) {
		if(!fh->is_floppy_present) {
			if(is_main_floppy_reload()) {
				fh->is_floppy_present = is_floppy_readable(fh);
			}
		}
		result = fh->is_floppy_present;
	} else if (fh->is_cdrom) {
		if(is_nt) {
			cache_clear( &fh->cache );
			result = true;
			/*
			DWORD dwBytesReturned;
			result = ( 0 != DeviceIoControl( 
				fh->h,
				IOCTL_STORAGE_CHECK_VERIFY,
				NULL, 0,
				NULL, 0,
				&dwBytesReturned,
				NULL
			) );
			*/
			if(result) {
				size_t bytes = 2048;
				char *buffer = (char *)VirtualAlloc( NULL, bytes, MEM_RESERVE|MEM_COMMIT, PAGE_READWRITE );
				if(buffer) {
					result = (Sys_read( arg, buffer, 0, 2048 ) == 2048);
					VirtualFree( buffer, 0, MEM_RELEASE  ); 
				} else {
					result = false;
				}
			}
		}
	}
	return(result);
}


/*
 *  Prevent medium removal (if applicable)
 */

void SysPreventRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;

	if (fh->is_cdrom) {
		if(is_nt) {
			PreventRemovalOfVolume(fh->h,true);
		}
	}
}


/*
 *  Allow medium removal (if applicable)
 */

void SysAllowRemoval(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh) return;

	if (fh->is_cdrom) {
		if(is_nt) {
			for( int i=0; i<10; i++) PreventRemovalOfVolume(fh->h,false);
		}
	}
}


/*
 *  Read CD-ROM TOC (binary MSF format, 804 bytes max.)
 */

bool SysCDReadTOC(void *arg, uint8 *toc)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;
}


/*
 *  Read CD-ROM position data (Sub-Q Channel, 16 bytes, see SCSI standard)
 */

bool SysCDGetPosition(void *arg, uint8 *pos)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;

}


/*
 *  Play CD audio
 */

bool SysCDPlay(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, uint8 end_m, uint8 end_s, uint8 end_f)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;
}


/*
 *  Pause CD audio
 */

bool SysCDPause(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;
}


/*
 *  Resume paused CD audio
 */

bool SysCDResume(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;
}


/*
 *  Stop CD audio
 */

bool SysCDStop(void *arg)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	return false;
}


/*
 *  Perform CD audio fast-forward/fast-reverse operation starting from specified address
 */

bool SysCDScan(void *arg, uint8 start_m, uint8 start_s, uint8 start_f, bool reverse)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return false;

	// Not supported under Linux
	return false;
}


/*
 *  Set CD audio volume (0..255 each channel)
 */

void SysCDSetVolume(void *arg, uint8 left, uint8 right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

}


/*
 *  Get CD audio volume (0..255 each channel)
 */

void SysCDGetVolume(void *arg, uint8 &left, uint8 &right)
{
	file_handle *fh = (file_handle *)arg;
	if (!fh)
		return;

	left = right = 0;
}
