/*
 *  video_windows.cpp - Video/graphics emulation for Win32
 *
 *  Basilisk II (C) 1997-1999 Christian Bauer
 */

#include "sysdeps.h"
#include "main.h"
#include "adb.h"
#include "prefs.h"
#include "user_strings.h"
#include "main_windows.h"
#include "video_windows.h"
#include "video.h"
#include <winnt.h>


/*
  If you undefine USE_FAULTS, apps like SpeedoMeter may give somewhat
  higher scores. Don't let this fool you, the overall performance is
  much better with USE_FAULTS defined, especially with high color modes
*/

int screen_bit_depth;

// should go to video.h
int32 MacFrameBaseDiff;

// Usually (but not always) it's best to use palette index
// and not explicit rgb values on 8bit devices
static bool bPalette = true;

#ifndef _DEBUG
#define USE_FAULTS
#endif

// Platform specific: size of the virtual memory page.
#define VM_PAGE_BYTES       4096
#define VM_PAGE_BITS        12
#define VM_PAGE_ALIGN_MASK  0xFFFFF000
#define VM_PAGE_MODULO_BITS 0xFFF


// Mac specific: screen size
#define SCREEN_BYTES    scalines[VideoMonitor.y]
#define SCREEN_MAX_X    VideoMonitor.x
#define SCREEN_MAX_Y    VideoMonitor.y
#define SCANLINE_BYTES  VideoMonitor.bytes_per_row

// How many virtual memory pages per Mac screen?.
// This just needs to be large enough. Should really be calculated
// using display attributes, fix it if you want
#define MAX_PAGES_PER_SCREEN 10000

#include "debug.h"


// Display types
enum {
  DISPLAY_WINDOW, // Win API
  DISPLAY_DX      // Direct X, not implemented
};

typedef struct {
  int top, bottom;  // Mapping between this virtual page and Mac scan lines
} ScreenPageInfo;

typedef struct {
  DWORD memBase,    // Real start address of this Mac screen page
        memStart,   // Start address aligned to nearest page boundary
        memEnd,     // Address of the first byte not belonging to the screen pages
        memLength;  // Length of the memory addressed by the screen pages
  int dirty,        // Anything written to this buffer?
      top,          // Top scanline of dirty area
      bottom;       // First "clean" scanline
  DWORD unprotect_start; // speedup.
  DWORD unprotect_bytes; // speedup
  ScreenPageInfo pages[MAX_PAGES_PER_SCREEN];
} ScreenInfo;


// Global variables
static int screen_inited = 0;

static int32 frame_skip = 1;
static uint8 *the_buffer = 0; 

static HBITMAP hScreenBitmap = 0;
static HDC hBitMapDC = 0;
static HDC hdcBits = 0;
static HPALETTE hPalMain = 0;
static HPALETTE hPalBits = 0;
static LPBITMAPINFO pbmi = 0;
static LPLOGPALETTE lppal = 0;

static HANDLE draw_mutex = 0;

#define MAX_SCAN_LINES 5000
int scalines[MAX_SCAN_LINES];


// Prototypes
DWORD WINAPI redraw_thread(LPVOID param);


static inline void grab_draw_mutex(void)
{
  WaitForSingleObject(draw_mutex,INFINITE);
}

static inline void release_draw_mutex(void)
{
  ReleaseMutex(draw_mutex);
}

#ifdef USE_FAULTS

static ScreenInfo mainBuffer;

// We don't really need this.
static LPTOP_LEVEL_EXCEPTION_FILTER nextFaultFunc = 0;

// This function is called when any exception would remain
// unhandled. It is a write to the Mac page if:
//    - it is a access violation
//    - it is within the range of the Mac screen buffers.
//    - if both of these are true, we know that it's a write
//      (since we didn't deny read).
// Update the dirty area (top,bottom) of the buffer,
// make the virtual page r/w again to stop faulting,
// and continue execution. The write will be attempted again
// and it will succeed now.

LONG WINAPI Screen_fault_proc( struct _EXCEPTION_POINTERS *ExceptionInfo )
{
  DWORD a;
  DWORD OldProtect;
  int page, lines, top, bottom;

  if(ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) {
		grab_draw_mutex();
    a = (DWORD)ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
    if(a >= mainBuffer.memStart && a < mainBuffer.memEnd) {
      page = (a - mainBuffer.memStart) >> VM_PAGE_BITS;
      top = mainBuffer.pages[page].top;
      bottom = mainBuffer.pages[page].bottom;

      lines = bottom-top;
      if(lines == 0) lines = 1;

      a = mainBuffer.memStart + scalines[top];

      if(mainBuffer.dirty) {
        mainBuffer.top    = min( mainBuffer.top, top );
        mainBuffer.bottom = max( mainBuffer.bottom, bottom );
        DWORD b = mainBuffer.unprotect_start+mainBuffer.unprotect_bytes;
        mainBuffer.unprotect_start = min(a,mainBuffer.unprotect_start);
        mainBuffer.unprotect_bytes = max(b,a+scalines[lines]) - mainBuffer.unprotect_start;
        VirtualProtect( (LPVOID)mainBuffer.unprotect_start, mainBuffer.unprotect_bytes, PAGE_READWRITE, &OldProtect );
      } else {
        mainBuffer.top    = top;
        mainBuffer.bottom = bottom;
        mainBuffer.dirty  = 1;
        mainBuffer.unprotect_start = a;
        mainBuffer.unprotect_bytes = scalines[lines];
        VirtualProtect( (LPVOID)a, mainBuffer.unprotect_bytes, PAGE_READWRITE, &OldProtect );
      }
			release_draw_mutex();
      return( EXCEPTION_CONTINUE_EXECUTION );
    }
  }
  return( EXCEPTION_CONTINUE_SEARCH );
}

void Screen_Setup_fault_handler(void)
{
  int i, y1, y2;
  DWORD a;

  mainBuffer.memBase  = (DWORD)the_buffer;
  mainBuffer.memStart = mainBuffer.memBase & VM_PAGE_ALIGN_MASK;
  mainBuffer.memEnd   = mainBuffer.memBase + SCREEN_BYTES;
  if(mainBuffer.memEnd & VM_PAGE_MODULO_BITS) mainBuffer.memEnd = (mainBuffer.memEnd | VM_PAGE_MODULO_BITS) + 1;
  mainBuffer.memLength = mainBuffer.memEnd - mainBuffer.memStart;
  mainBuffer.dirty = 1;
  mainBuffer.top = 0;
  mainBuffer.bottom = SCREEN_MAX_Y;

  mainBuffer.unprotect_start = mainBuffer.memStart;
  mainBuffer.unprotect_bytes = mainBuffer.memLength;

  a = 0;
  for(i=0; i<(unsigned int)MAX_PAGES_PER_SCREEN; i++) {
    y1 = a / (int)SCANLINE_BYTES;
    y2 = (a+VM_PAGE_BYTES) / (int)SCANLINE_BYTES;

    y1 = y1 - 8;
    y2 = y2 + 32;
		
    // y1 = y1 - 16;
    // y2 = y2 + 64;
		
    if(y1 < 0) y1 = 0;
    if(y2 < 0) y2 = 0;
    if(y1 > (int)VideoMonitor.y) y1 = (int)VideoMonitor.y;
    if(y2 > (int)VideoMonitor.y) y2 = (int)VideoMonitor.y;

    mainBuffer.pages[i].top = y1;
    mainBuffer.pages[i].bottom = y2;
    a += VM_PAGE_BYTES;
  }

  nextFaultFunc = SetUnhandledExceptionFilter( (LPTOP_LEVEL_EXCEPTION_FILTER)Screen_fault_proc );
}
#endif //USE_FAULTS

/*
 *  Initialization
 */

// Set VideoMonitor according to video mode
void set_video_monitor(int width, int height, int depth, int bytes_per_row)
{
	screen_bit_depth = depth;

  switch (depth) {
    case 8:
      MacFrameLayout = FLAYOUT_DIRECT;
      VideoMonitor.mode = VMODE_8BIT;
      break;
    case 15:
      MacFrameLayout = FLAYOUT_HOST_555;
      VideoMonitor.mode = VMODE_16BIT;
      break;
    case 16:
      MacFrameLayout = FLAYOUT_HOST_565;
      VideoMonitor.mode = VMODE_16BIT;
      break;
    case 24:
      MacFrameLayout = FLAYOUT_HOST_888;
      VideoMonitor.mode = VMODE_32BIT;
      break;
    case 32:
      MacFrameLayout = FLAYOUT_HOST_888;
      VideoMonitor.mode = VMODE_32BIT;
      break;
  }
  VideoMonitor.x = width;
  VideoMonitor.y = height;
  VideoMonitor.bytes_per_row = bytes_per_row;

	for( int i=0; i<MAX_SCAN_LINES; i++ ) {
		scalines[i] = bytes_per_row * i;
	}
}

static int calc_bytes_per_row( int width, int depth )
{
  int result;
  depth = ((depth + 7) / 8) * 8;
  result = width * depth / 8;
  return(result);
}

// Init window mode
static bool init_window(int width, int height, int depth_mac, int depth_win)
{
  bool retval = false;

  // This doesn't belong here
  ADBSetRelMouseMode(false);

  frame_skip = PrefsFindInt32("frameskip");
  if (frame_skip <= 0) frame_skip = 1;

  set_video_monitor(width, height, depth_win, calc_bytes_per_row(width,depth_mac) );
  the_buffer = (uint8 *)VirtualAlloc( 0, width * calc_bytes_per_row(height,depth_mac) + VM_PAGE_BYTES*16, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );

#if REAL_ADDRESSING
  VideoMonitor.mac_frame_base = (uint32)the_buffer;
  MacFrameLayout = FLAYOUT_DIRECT;
#else
  VideoMonitor.mac_frame_base = MacFrameBaseMac;
#endif

  HDC hScreenDC = GetDC(hMainWnd);
  if(hScreenDC) {
    hScreenBitmap = CreateBitmap(VideoMonitor.x, VideoMonitor.y, 1, depth_win, NULL);
    hdcBits = CreateCompatibleDC(hScreenDC);
    ReleaseDC(hMainWnd,hScreenDC);
  }

  lppal = (LPLOGPALETTE)malloc( sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * 256 );
  lppal->palVersion = 0x300;
  lppal->palNumEntries = (WORD)256;

  retval = (hScreenBitmap && hdcBits && the_buffer && lppal);

  if(!retval) {
    if(hScreenBitmap) {
      DeleteObject(hScreenBitmap);
      hScreenBitmap = 0;
    }
    if(hdcBits) {
      DeleteDC(hdcBits);
      hdcBits = 0;
    }
		if(lppal) {
			free(lppal);
			lppal = 0;
		}
  }

  return ( retval );
}

// Init DX display
static bool init_dx(int width, int height, int depth_mac, int depth_win)
{
  ADBSetRelMouseMode(true);
  set_video_monitor(width, height, depth_win, calc_bytes_per_row(width,depth_mac) );
#if REAL_ADDRESSING
  VideoMonitor.mac_frame_base = (uint32)the_buffer;
  MacFrameLayout = FLAYOUT_DIRECT;
#else
  VideoMonitor.mac_frame_base = MacFrameBaseMac;
#endif
  return true;
}

bool VideoInit(void)
{
  // Enable thread support
  int display_type = DISPLAY_WINDOW;
  int width, height, depth_mac, depth_win;

  if(screen_inited) return 0;

  draw_mutex = CreateMutex( 0, FALSE, "Basilisk II draw mutex" );
  if(draw_mutex == 0) {
    ErrorAlert("Cannot create Basilisk II draw mutex.");
    QuitEmulator();
  }

  // Determine type and mode

  RECT rc;
  GetClientRect( hMainWnd, &rc );
  width = rc.right;
  height = rc.bottom;

  HDC hScreenDC = GetDC(hMainWnd);
	if(hScreenDC) {
	  depth_mac = depth_win = GetDeviceCaps(hScreenDC,BITSPIXEL);
		ReleaseDC(hMainWnd,hScreenDC);
	}

  if(depth_mac == 24) depth_mac = 32;

  /*
  const char *mode_str = PrefsFindString("screen");
  if (mode_str) {
    if (sscanf(mode_str, "win/%d/%d", &width, &height) == 2)
      display_type = DISPLAY_WINDOW;
    else if (strcmp(mode_str, "dx") == 0) {
      display_type = DISPLAY_DX;
    }
  }
  */

  // This is used only in 8 bit palette mode
  pbmi = (LPBITMAPINFO)malloc( sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );
  
  if(!pbmi) exit(0); // impossible..

  memset( pbmi, 0, sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );

  pbmi = (LPBITMAPINFO)malloc( sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD) );
  pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  pbmi->bmiHeader.biWidth = width;
  pbmi->bmiHeader.biPlanes = 1;
  pbmi->bmiHeader.biBitCount = depth_mac;
  pbmi->bmiHeader.biCompression = 0;
  pbmi->bmiHeader.biXPelsPerMeter = 0;
  pbmi->bmiHeader.biYPelsPerMeter = 0;
  pbmi->bmiHeader.biClrUsed = depth_mac == 8 ? 256 : 0;
  pbmi->bmiHeader.biClrImportant = depth_mac == 8 ? 256 : 0;

	for( int i=0; i<256; i++ ) {
    pbmi->bmiColors[i].rgbReserved = 0; //PC_NOCOLLAPSE|PC_RESERVED;
	}

  // Initialize according to display type
  switch (display_type) {
    case DISPLAY_WINDOW:
      if (!init_window(width, height, depth_mac, depth_win))
        return false;
      break;
    case DISPLAY_DX:
      // not reached.
      if (!init_dx(width, height, depth_mac, depth_win))
        return false;
      break;
  }

#ifdef USE_FAULTS
  Screen_Setup_fault_handler();
#endif //USE_FAULTS

#if !REAL_ADDRESSING
  // Set variables for UAE memory mapping
  MacFrameBaseHost = the_buffer;
  MacFrameSize = VideoMonitor.bytes_per_row * VideoMonitor.y;
#endif

	MacFrameBaseDiff = (int32)MacFrameBaseHost - (int32)MacFrameBaseMac;

  screen_inited = 1;

	DWORD tid;
  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)redraw_thread, (LPVOID) hMainWnd, 0, &tid);

  return true;
}


/*
 *  Deinitialization
 */

void VideoExit(void)
{
  screen_inited = 0;
  if(the_buffer) {
    VirtualFree( the_buffer, 0, MEM_RELEASE  ); 
  }
  if(lppal) {
		free(lppal);
		lppal = 0;
	}
  if(hPalMain) {
    DeleteObject(hPalMain);
    hPalMain = 0;
  }
  if(hPalBits) {
    DeleteObject(hPalBits);
    hPalBits = 0;
  }
  if(pbmi) {
    free(pbmi);
    pbmi = 0;
  }
  if(draw_mutex) {
    CloseHandle(draw_mutex);
    draw_mutex = 0;
  }
  if(hScreenBitmap) {
    DeleteObject(hScreenBitmap);
    hScreenBitmap = 0;
  }
  if(hdcBits) {
    DeleteDC(hdcBits);
    hdcBits = 0;
  }
}

#pragma pack(1)
typedef struct {
	BYTE red, green, blue;
} macpal_type;
#pragma pack()

typedef struct {
	int32 diff;
	uint32 inx;
} macpal_inx;

int __cdecl palsort( const void *a, const void *b)
{
	return( ((macpal_inx *)b)->diff - ((macpal_inx *)a)->diff );
}

void setup_palette_entries( HDC hdc, macpal_type pal[256])
{
	int i;
	macpal_inx diff[256];
	BYTE red, green, blue;

	diff[0].inx = 0;
	diff[0].diff = 50000;
	diff[255].inx = 255;
	diff[255].diff = 50000;

	GetSystemPaletteEntries( hdc, 0, 20, &lppal->palPalEntry[20] );

  for( i=1; i<255; i++ ) {
		diff[i].inx = i;
		red = pal[i].red;
		green = pal[i].green;
		blue = pal[i].blue;
		if( 0 && (red == 0 || green == 0 || blue == 0 || 
		   red == 255 || green == 255 || blue == 255) )
		{
			diff[i].diff = 50000;
		} else {
			diff[i].diff = 
				abs( (int)pal[i-1].red - (int)pal[i].red) + 
				abs( (int)pal[i-1].green - (int)pal[i].green) + 
				abs( (int)pal[i-1].blue - (int)pal[i].blue);
		}
	}
	qsort( (void *)diff, 256, sizeof(diff[0]), palsort );

  for (i=0; i<236; i++) {
    lppal->palPalEntry[i+20].peRed = pal[diff[i].inx].red;
    lppal->palPalEntry[i+20].peGreen = pal[diff[i].inx].green;
    lppal->palPalEntry[i+20].peBlue = pal[diff[i].inx].blue;
    lppal->palPalEntry[i+20].peFlags = PC_NOCOLLAPSE;
    pbmi->bmiColors[i].rgbReserved = 0; // PC_RESERVED|PC_NOCOLLAPSE
  }
}

void setup_palette_entries(uint8 *pal)
{
	uint16 i, j;
	uint16 *iptr = (uint16 *)&pbmi->bmiColors[0];

  for (i=0,j=0; i<256; i++,j+=3) {
    lppal->palPalEntry[i].peRed = pal[j];
    lppal->palPalEntry[i].peGreen = pal[j+1];
    lppal->palPalEntry[i].peBlue = pal[j+2];
    lppal->palPalEntry[i].peFlags = PC_NOCOLLAPSE; // PC_RESERVED|PC_NOCOLLAPSE
		if(bPalette) {
			*iptr++ = i;
		} else {
			pbmi->bmiColors[i].rgbRed = pal[j];
			pbmi->bmiColors[i].rgbGreen = pal[j+1];
			pbmi->bmiColors[i].rgbBlue = pal[j+2];
			pbmi->bmiColors[i].rgbReserved = 0;
		}
  }
}

/*
 *  Set palette
 */

void video_set_palette(uint8 *pal)
{
	grab_draw_mutex();

  if(screen_bit_depth == 8) {
		HDC hScreenDC = GetDC(hMainWnd);
		if(hScreenDC) {
			HPALETTE hold;

			if(hPalMain) DeleteObject(hPalMain);
			if(hPalBits) DeleteObject(hPalBits);

			setup_palette_entries( pal );
			hPalMain = CreatePalette(lppal);
			hold = SelectPalette( hScreenDC, hPalMain, FALSE );
			RealizePalette(hScreenDC);
			SelectPalette( hScreenDC, hold, FALSE );

			hPalBits = CreatePalette(lppal);
			hold = SelectPalette( hdcBits, hPalBits, FALSE );
			RealizePalette(hdcBits);
			SelectPalette( hdcBits, hold, FALSE );
			ReleaseDC(hMainWnd,hScreenDC);
	  }
  }
  
#ifdef USE_FAULTS
  mainBuffer.dirty = 1;
  mainBuffer.top = 0;
  mainBuffer.bottom = SCREEN_MAX_Y;
  mainBuffer.unprotect_start = mainBuffer.memStart;
  mainBuffer.unprotect_bytes = mainBuffer.memLength;
#endif

	release_draw_mutex();
}

int update_palette()
{
	int i = 0;

  if(hPalMain && hPalBits) {
		grab_draw_mutex();
    HDC hScreenDC = GetDC(hMainWnd);
		if(hScreenDC) {
			HPALETTE hold;
			UnrealizeObject(hPalMain);
			UnrealizeObject(hPalBits);
		  hold = SelectPalette( hScreenDC, hPalMain, FALSE );
	    i = RealizePalette(hScreenDC);
			SelectPalette( hScreenDC, hold, FALSE );
	    hold = SelectPalette( hdcBits, hPalBits, FALSE );
	    i += RealizePalette(hdcBits);
			SelectPalette( hdcBits, hold, FALSE );
			if(i) {
				InvalidateRect( hMainWnd, NULL, FALSE );
#ifdef USE_FAULTS
				mainBuffer.dirty = 1;
				mainBuffer.top = 0;
				mainBuffer.bottom = SCREEN_MAX_Y;
				mainBuffer.unprotect_start = mainBuffer.memStart;
				mainBuffer.unprotect_bytes = mainBuffer.memLength;
#endif
			}
	    ReleaseDC(hMainWnd,hScreenDC);
		}
		release_draw_mutex();
  }
	return( i != 0 );
}

#ifdef USE_FAULTS
void Screen_Draw(HDC hScreenDC)
{
  BYTE *pbyScreen;
  DWORD OldProtect;
  HBITMAP oldBitmap;
  int y, line_count;
	HPALETTE hold1, hold2;

  if(screen_inited && mainBuffer.dirty) {
    mainBuffer.dirty = 0;

    VirtualProtect( (LPVOID)mainBuffer.unprotect_start, mainBuffer.unprotect_bytes, PAGE_READONLY, &OldProtect );

    y = mainBuffer.top;
		line_count = mainBuffer.bottom - y;
    pbyScreen = &((BYTE *)mainBuffer.memBase)[scalines[y]];

		switch(screen_bit_depth) {
			case 8:
				hold1 = SelectPalette( hScreenDC, hPalMain, FALSE );
				hold2 = SelectPalette( hdcBits, hPalBits, FALSE );
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = scalines[line_count];
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, 
					bPalette ? DIB_PAL_COLORS : DIB_RGB_COLORS );
				break;
			case 15:
				// does not work
				SetBitmapBits(hScreenBitmap, scalines[line_count], pbyScreen );
				// pbmi->bmiHeader.biHeight = -line_count;
				// pbmi->bmiHeader.biSizeImage = scalines[line_count];
				// SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
			case 16:
				// SetDIBits version is faster but the colors are not correct
				SetBitmapBits(hScreenBitmap, scalines[line_count], pbyScreen );
				// pbmi->bmiHeader.biHeight = -line_count;
				// pbmi->bmiHeader.biSizeImage = scalines[line_count];
				// SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
			case 24:
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = (line_count*VideoMonitor.x) << 2;
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
			case 32:
				SetBitmapBits(hScreenBitmap, scalines[line_count], pbyScreen );
				pbmi->bmiHeader.biHeight = -line_count;
				pbmi->bmiHeader.biSizeImage = scalines[line_count];
				SetDIBits( hdcBits, hScreenBitmap, 0, line_count, pbyScreen, pbmi, DIB_RGB_COLORS );
				break;
		}

    oldBitmap = SelectObject(hdcBits,hScreenBitmap);
    BitBlt(hScreenDC, 0, y, VideoMonitor.x, line_count, hdcBits, 0, 0, SRCCOPY );
		if(hPalMain) {
			SelectPalette( hScreenDC, hold1, FALSE );
			SelectPalette( hdcBits, hold2, FALSE );
		}
    SelectObject(hdcBits,oldBitmap);
  }
}

void Screen_Draw_All(HWND hWnd)
{
  HDC hdc;
  PAINTSTRUCT ps;

	grab_draw_mutex();
  hdc = BeginPaint(hWnd, &ps);
  mainBuffer.dirty = 1;
  mainBuffer.top = 0;
  mainBuffer.bottom = SCREEN_MAX_Y;
  mainBuffer.unprotect_start = mainBuffer.memStart;
  mainBuffer.unprotect_bytes = mainBuffer.memLength;
  Screen_Draw(hdc);
  EndPaint(hWnd, &ps);
	release_draw_mutex();
}
#else
void Screen_Draw(HDC hScreenDC)
{
	HPALETTE hold1, hold2;

  if(!screen_inited) return;

	// This code is for debugging only and not necessarily up to date.
	// See the release code above 

	switch(screen_bit_depth) {
		case 8:
			hold1 = SelectPalette( hScreenDC, hPalMain, FALSE );
			hold2 = SelectPalette( hdcBits, hPalBits, FALSE );
			pbmi->bmiHeader.biHeight = -(int)VideoMonitor.y;
			pbmi->bmiHeader.biSizeImage = scalines[VideoMonitor.y];
			SetDIBits( hdcBits, hScreenBitmap, 0, VideoMonitor.y, the_buffer, pbmi, 
					bPalette ? DIB_PAL_COLORS : DIB_RGB_COLORS );
			break;
		case 24:
			pbmi->bmiHeader.biHeight = -(int)VideoMonitor.y;
			pbmi->bmiHeader.biSizeImage = VideoMonitor.y*VideoMonitor.x*4;
			SetDIBits( hdcBits, hScreenBitmap, 0, VideoMonitor.y, the_buffer, pbmi, DIB_RGB_COLORS );
			break;
		default:
	    SetBitmapBits(hScreenBitmap, scalines[VideoMonitor.y], the_buffer );
	}

  HBITMAP oldBitmap = SelectObject(hdcBits,hScreenBitmap);
  BitBlt(hScreenDC, 0, 0, VideoMonitor.x, VideoMonitor.y, hdcBits, 0, 0, SRCCOPY );
	if(hPalMain) {
		SelectPalette( hScreenDC, hold1, FALSE );
		SelectPalette( hdcBits, hold2, FALSE );
	}
  SelectObject(hdcBits,oldBitmap);
}

void Screen_Draw_All(HWND hWnd)
{
  HDC hdc;
  PAINTSTRUCT ps;

	grab_draw_mutex();
  hdc = BeginPaint(hWnd, &ps);
  Screen_Draw(hdc);
  EndPaint(hWnd, &ps);
	release_draw_mutex();
}
#endif //USE_FAULTS

/*
 *  Thread for screen refresh
 */

DWORD WINAPI redraw_thread(LPVOID param)
{
#ifndef USE_FAULTS
  int tick_counter = 0;
#endif
  HDC hScreenDC;

  for (;;) {
		Sleep(40);
#ifndef USE_FAULTS
    if (++tick_counter >= frame_skip) {
      tick_counter = 0;
#endif
			grab_draw_mutex();
      hScreenDC = GetDC(hMainWnd);
      if(hScreenDC) {
        Screen_Draw(hScreenDC);
        ReleaseDC(hMainWnd,hScreenDC);
      }
			release_draw_mutex();
#ifndef USE_FAULTS
    }
#endif
  }
  return 0;
}

void VideoInterrupt(void)
{
}
